# AUTOGENERATED! DO NOT EDIT! File to edit: ../nbs/API/co.ipynb.

# %% auto 0
__all__ = ['emperical_co', 'emperical_co_pc', 'isPD', 'nearestPD', 'regularize_spectral']

# %% ../nbs/API/co.ipynb 4
import cupy as cp
from typing import Union

# %% ../nbs/API/co.ipynb 6
_emperical_co_kernel = cp.ElementwiseKernel(
    'raw T rslc, raw bool is_shp, int32 nlines, int32 width, int32 nimages, int32 az_half_win, int32 r_half_win',
    'raw T cov, raw T coh',
    '''
    if (i >= nlines*width) return;
    int az_win = 2*az_half_win+1;
    int r_win = 2*r_half_win+1;
    int win = az_win*r_win;
    
    int ref_az = i/width;
    int ref_r = i -ref_az*width;

    int sec_az, sec_r;

    int m,j; // index of each coherence matrix
    int k,l; // index of search window
    T _cov; // covariance
    float _amp2_m; // sum of amplitude square for image i
    float _amp2_j; // sum of amplitude aquare for image j
    int rslc_inx_m, rslc_inx_j;
    int n; // number of shp

    for (m = 0; m < nimages; m++) {
        for (j = 0; j < nimages; j++) {
            _cov = T(0.0, 0.0);
            _amp2_m = 0.0;
            _amp2_j = 0.0;
            n = 0;
            for (k = 0; k < az_win; k++) {
                for (l = 0; l < r_win; l++) {
                    sec_az = ref_az-az_half_win+k;
                    sec_r = ref_r-r_half_win+l;
                    if (is_shp[i*win+k*r_win+l] && sec_az >= 0 && sec_az < nlines && sec_r >= 0 && sec_r < width) {
                        rslc_inx_m = (sec_az*width+sec_r)*nimages+m;
                        rslc_inx_j = (sec_az*width+sec_r)*nimages+j;
                        _amp2_m += norm(rslc[rslc_inx_m]);
                        _amp2_j += norm(rslc[rslc_inx_j]);
                        _cov += rslc[rslc_inx_m]*conj(rslc[rslc_inx_j]);
                        n += 1;
                        //if (i == 0 && m ==3 && j == 1) {
                        //    printf("%f",_cov.real());
                        //}
                    }
                }
            }
            cov[(i*nimages+m)*nimages+j] = _cov/(float)n;
            //if ( i == 0 && m==3 && j ==1 ) printf("%d",((i*nimages+m)*nimages+j));
            _amp2_m = sqrt(_amp2_m*_amp2_j);
            coh[(i*nimages+m)*nimages+j] = _cov/_amp2_m;
        }
    }
    ''',
    name = 'emperical_co_kernel',reduce_dims = False,no_return=True
)

# %% ../nbs/API/co.ipynb 8
def emperical_co(rslc:cp.ndarray, # rslc stack, dtype: `cupy.complexfloating`
                 is_shp:cp.ndarray, # shp bool, dtype: `cupy.bool`
                 block_size:int=128, # the CUDA block size, it only affects the calculation speed
                )-> tuple[cp.ndarray,cp.ndarray]: # the covariance and coherence matrix `cov` and `coh`
    '''
    Maximum likelihood covariance estimator.
    '''
    nlines, width, nimages = rslc.shape
    az_win, r_win = is_shp.shape[-2:]
    az_half_win = (az_win-1)//2
    r_half_win = (r_win-1)//2

    cov = cp.empty((nlines,width,nimages,nimages),dtype=rslc.dtype)
    coh = cp.empty((nlines,width,nimages,nimages),dtype=rslc.dtype)

    _emperical_co_kernel(rslc, is_shp, cp.int32(nlines),cp.int32(width),cp.int32(nimages),
                    cp.int32(az_half_win),cp.int32(r_half_win),cov,coh,size = nlines*width,block_size=block_size)
    return cov,coh

# %% ../nbs/API/co.ipynb 17
# I is int32* or int64*
_emperical_co_pc_kernel = cp.ElementwiseKernel(
    'raw T rslc, raw I az_idx, raw I r_idx, raw bool pc_is_shp, int32 nlines, int32 width, int32 nimages, int32 az_half_win, int32 r_half_win, int32 n_pc',
    'raw T cov, raw T coh',
    '''
    if (i >= n_pc) return;
    int az_win = 2*az_half_win+1;
    int r_win = 2*r_half_win+1;
    int win = az_win*r_win;
    
    int ref_az = az_idx[i];
    int ref_r = r_idx[i];

    int sec_az, sec_r;

    int m,j; // index of each coherence matrix
    int k,l; // index of search window
    T _cov; // covariance
    float _amp2_m; // sum of amplitude square for image i
    float _amp2_j; // sum of amplitude aquare for image j
    int rslc_inx_m, rslc_inx_j;
    int n; // number of shp

    for (m = 0; m < nimages; m++) {
        for (j = 0; j < nimages; j++) {
            _cov = T(0.0, 0.0);
            _amp2_m = 0.0;
            _amp2_j = 0.0;
            n = 0;
            for (k = 0; k < az_win; k++) {
                for (l = 0; l < r_win; l++) {
                    sec_az = ref_az-az_half_win+k;
                    sec_r = ref_r-r_half_win+l;
                    if (pc_is_shp[i*win+k*r_win+l] && sec_az >= 0 && sec_az < nlines && sec_r >= 0 && sec_r < width) {
                        rslc_inx_m = (sec_az*width+sec_r)*nimages+m;
                        rslc_inx_j = (sec_az*width+sec_r)*nimages+j;
                        _amp2_m += norm(rslc[rslc_inx_m]);
                        _amp2_j += norm(rslc[rslc_inx_j]);
                        _cov += rslc[rslc_inx_m]*conj(rslc[rslc_inx_j]);
                        n += 1;
                        //if (i == 0 && m ==3 && j == 1) {
                        //    printf("%f",_cov.real());
                        //}
                    }
                }
            }
            cov[(i*nimages+m)*nimages+j] = _cov/(float)n;
            //if ( i == 0 && m==3 && j ==1 ) printf("%d",((i*nimages+m)*nimages+j));
            _amp2_m = sqrt(_amp2_m*_amp2_j);
            coh[(i*nimages+m)*nimages+j] = _cov/_amp2_m;
        }
    }
    ''',
    name = 'emperical_co_pc_kernel',reduce_dims = False,no_return=True
)

# %% ../nbs/API/co.ipynb 18
def emperical_co_pc(rslc:cp.ndarray, # rslc stack, dtype: `cupy.complexfloating`
                    idx:cp.ndarray, # index of point target (azimuth_index, range_index), dtype: `cupy.int`, shape: (2,n_sp)
                    pc_is_shp:cp.ndarray, # shp bool, dtype: `cupy.bool`
                    block_size:int=128, # the CUDA block size, it only affects the calculation speed
                   )-> tuple[cp.ndarray,cp.ndarray]: # the covariance and coherence matrix `cov` and `coh`
    '''
    Maximum likelihood covariance estimator for sparse data.
    '''
    nlines, width, nimages = rslc.shape
    az_win, r_win = pc_is_shp.shape[-2:]
    az_half_win = (az_win-1)//2
    r_half_win = (r_win-1)//2
    az_idx = idx[0]; r_idx = idx[1]
    n_pc = az_idx.shape[0]

    
    cov = cp.empty((n_pc,nimages,nimages),dtype=rslc.dtype)
    coh = cp.empty((n_pc,nimages,nimages),dtype=rslc.dtype)

    _emperical_co_pc_kernel(rslc, az_idx, r_idx, pc_is_shp, cp.int32(nlines),cp.int32(width),cp.int32(nimages),
                    cp.int32(az_half_win),cp.int32(r_half_win),cp.int32(n_pc),cov,coh,size = n_pc,block_size=block_size)
    return cov,coh

# %% ../nbs/API/co.ipynb 26
def isPD(co:cp.ndarray, # absolute value of complex coherence/covariance stack
         )-> cp.ndarray: # bool array indicating wheather coherence/covariance is positive define
    L = cp.linalg.cholesky(co)
    is_PD = cp.isfinite(L).all(axis=(-2,-1))
    return is_PD

# %% ../nbs/API/co.ipynb 33
'''
    The method is presented in [1]. John D'Errico implented it in MATLAB [2] under BSD
    Licence and [3] implented it with Python/Numpy based on [2] also under BSD Licence.
    This is a cupy implentation with stack of matrix supported.

    [1] N.J. Higham, "Computing a nearest symmetric positive semidefinite
    matrix" (1988): https://doi.org/10.1016/0024-3795(88)90223-6
    
    [2] https://www.mathworks.com/matlabcentral/fileexchange/42885-nearestspd
    
    [3] https://gist.github.com/fasiha/fdb5cec2054e6f1c6ae35476045a0bbd
'''
def nearestPD(co:cp.ndarray, # stack of matrix with shape [...,N,N]
             )-> cp.ndarray: # nearest positive definite matrix of input, shape [...,N,N]
    """Find the nearest positive-definite matrix to input matrix."""

    B = (co + cp.swapaxes(co,-1,-2))/2
    s, V = cp.linalg.svd(co)[1:]
    I = cp.eye(co.shape[-1],dtype=co.dtype)
    S = s[...,None]*I
    del s

    H = cp.matmul(cp.swapaxes(V,-1,-2), cp.matmul(S, V))
    del S, V
    A2 = (B + H) / 2
    del B, H
    A3 = (A2 + cp.swapaxes(A2,-1,-2))/2
    del A2

    if wherePD(A3).all():
        return A3
    
    co_norm = cp.linalg.norm(co,axis=(-2,-1))
    spacing = cp.nextafter(co_norm,co_norm+1.0)-co_norm
    
    k = 0
    while True:
        isPD = wherePD(A3)
        isPD_all = isPD.all()
        if isPD_all or k>=100:
            break
        k+=1
        mineig = cp.amin(cp.linalg.eigvalsh(A3),axis=-1)
        assert cp.isfinite(mineig).all()
        A3 += (~isPD[...,None,None] * I) * (-mineig * k**2 + spacing)[...,None,None]
    #print(k)
    return A3

# %% ../nbs/API/co.ipynb 35
def regularize_spectral(coh:cp.ndarray, # stack of matrix with shape [...,N,N]
                        beta:Union[float, cp.ndarray], # the regularization parameter, a float number or cupy ndarray with shape [...]
                        )-> cp.ndarray: # regularized matrix, shape [...,N,N]
    '''
    Spectral regularizer for coherence matrix.
    '''
    I = cp.eye(coh.shape[-1],dtype=coh.dtype)
    beta = cp.asarray(beta)[...,None,None]

    regularized_coh = (1-beta)*coh + beta* I
    return regularized_coh



# AUTOGENERATED! DO NOT EDIT! File to edit: ../nbs/API/plot.ipynb.

# %% auto 0
__all__ = ['bg_alpha']

# %% ../nbs/API/plot.ipynb 3
import numpy as np

# %% ../nbs/API/plot.ipynb 4
def bg_alpha(pwr):
    _pwr = np.power(pwr,0.35)
    cv = _pwr.mean()*2.5
    v = (_pwr.clip(0., cv))/cv
    return v



# AUTOGENERATED! DO NOT EDIT! File to edit: ../nbs/API/pl.ipynb.

# %% auto 0
__all__ = ['emi', 'temp_coh']

# %% ../nbs/API/pl.ipynb 5
import cupy as cp

# %% ../nbs/API/pl.ipynb 8
def emi(coh:cp.ndarray, #complex coherence metrix,dtype cupy.complex
        ref:int=0, #index of reference image in the phase history output, optional. Default: 0
       )-> tuple[cp.ndarray,cp.ndarray]: # estimated phase history `ph`, dtype complex; quality (minimum eigvalue, dtype float)
    coh_mag = abs(coh)
    coh_mag_inv = cp.linalg.inv(coh_mag)
    min_eigval, min_eig = cp.linalg.eigh(coh_mag_inv*coh)
    min_eigval = min_eigval[...,0]
    # min_eig = min_eig[...,0]
    min_eig = min_eig[...,0]*min_eig[...,[ref],0].conj()

    return min_eig/abs(min_eig), min_eigval

# %% ../nbs/API/pl.ipynb 17
def temp_coh(coh:cp.ndarray,# complex coherence metrix, dtype cupy.complex
             ds_ph = cp.ndarray, # complex phase history, dtype cupy.complex
            ):
    nimages = ds_ph.shape[-1]
    assert coh.shape[-2:] == (nimages,nimages), "input dimension not match"
    ds_ph = ds_ph/abs(ds_ph)
    coh = coh/abs(coh)
    int_ph = ds_ph[...,None]*ds_ph[...,None,:].conj()
    diff_ph = coh*int_ph.conj()
    t_coh = (cp.sum(diff_ph,axis=(-2,-1))-cp.trace(diff_ph,axis1=-2,axis2=-1)).real/(nimages**2-nimages)
    return t_coh



# AUTOGENERATED! DO NOT EDIT! File to edit: ../../../nbs/CLI/utils/appconfig.ipynb.

# %% auto 0
__all__ = ['get_config']

# %% ../../../nbs/CLI/utils/appconfig.ipynb 3
from pathlib import Path
import toml

# %% ../../../nbs/CLI/utils/appconfig.ipynb 4
def get_config(appname = 'decorrelation'):
    files = [Path(f'~/.{appname}.toml').expanduser(), Path(f'./{appname}.toml')]
    config_dict = dict()
    for file in files:
        if file.exists():
            config_dict.update(toml.load(file))
    return config_dict



# AUTOGENERATED! DO NOT EDIT! File to edit: ../../nbs/CLI/load.ipynb.

# %% auto 0
__all__ = ['read_gamma_image', 'write_gamma_image', 'read_gamma_plist', 'write_gamma_plist', 'de_load_gamma_flatten_rslc',
           'de_load_gamma_lat_lon_hgt', 'de_load_gamma_look_vector', 'de_load_gamma_range', 'de_load_gamma_metadata']

# %% ../../nbs/CLI/load.ipynb 4
import glob
from pathlib import Path
import tempfile
import re
import os
import toml
import zarr

import numpy as np
from matplotlib import pyplot as plt
import pandas as pd
from scipy.constants import speed_of_light
import dask
from dask import array as da
from dask import delayed
from dask.distributed import Client, LocalCluster

from .utils.logging import get_logger, log_args

from fastcore.script import call_parse

# %% ../../nbs/CLI/load.ipynb 5
def _rdc_width_nlines(image_par):
    """get slc width and number of lines.
    """
    with open(image_par) as f:
        for line in f:
            if re.search('range_samples',line):
                rdc_width = int(line.split()[1])
            if re.search('azimuth_lines',line):
                rdc_nlines = int(line.split()[1])
    return rdc_width, rdc_nlines

# %% ../../nbs/CLI/load.ipynb 6
def _geo_width_nlines(dem_par):
    """get dem width and number of lines.
    """
    with open(dem_par) as f:
        for line in f:
            if re.search('width',line):
                geo_width = int(line.split()[1])
            if re.search('nlines',line):
                geo_nlines = int(line.split()[1])
    return geo_width, geo_nlines

# %% ../../nbs/CLI/load.ipynb 7
def _cor_pos_dem(dem_par):
    """get corner lat and lon and post lat and lon
    """
    with open(dem_par) as f:
        for line in f:
            if re.search('corner_lat',line):
                cor_lat = float(line.split()[1])
            if re.search('corner_lon',line):
                cor_lon = float(line.split()[1])
            if re.search('post_lat',line):
                pos_lat = float(line.split()[1])
            if re.search('post_lon',line):
                pos_lon = float(line.split()[1])
    return cor_lat, cor_lon, pos_lat, pos_lon

# %% ../../nbs/CLI/load.ipynb 8
def _fetch_slc_par_date(rslc_dir,# str / Path
                       ):
    rslc_dir = Path(rslc_dir)
    rslcs = []
    rslc_pars = []
    dates = []
    for rslc in sorted(rslc_dir.glob('*.rslc')):
        rslc_par = rslc.parent / (rslc.name + '.par')
        assert rslc_par.exists(), f'{str(rslc_par)} not exists!'
        date = rslc.stem

        rslcs.append(rslc)
        rslc_pars.append(rslc_par)
        dates.append(date)
    rslcs_df = pd.DataFrame({'date':dates,'rslc':rslcs,'par':rslc_pars})
    return rslcs_df

# %% ../../nbs/CLI/load.ipynb 9
def read_gamma_image(imag:str, # gamma raster data
                     width:int, # data width
                     dtype:str='float', # data format, only 'float' and 'fcomplex' are supported
                    ):
    '''read gamma image into numpy array.'''
    if dtype == 'float':
        dt = '>f4'
    elif dtype == 'fcomplex':
        dt = '>c8'
    elif dtype == 'int':
        dt = '>i4'
    elif dtype =='double':
        dt = '>f8'
    else: raise ValueError('Unsupported data type')

    datf = open(imag,"r")
    datf.seek(0,os.SEEK_END)
    size = datf.tell()
    nlines = int(size / int(dt[-1]) / width)
    data = np.fromfile(imag,dtype=dt,count=width*nlines)
    return data.reshape(-1,width)

# %% ../../nbs/CLI/load.ipynb 10
def write_gamma_image(imag,path):
    imag = imag.astype(imag.dtype.newbyteorder('big'))
    imag.tofile(path)

# %% ../../nbs/CLI/load.ipynb 11
def read_gamma_plist(plist:str,dtype='int'):
    return read_gamma_image(plist,width=2,dtype=dtype)

# %% ../../nbs/CLI/load.ipynb 12
def write_gamma_plist(imag,path):
    return write_gamma_image(imag,path)

# %% ../../nbs/CLI/load.ipynb 14
@call_parse
@log_args
def de_load_gamma_flatten_rslc(rslc_dir:str, # gamma rslc directory, the name of the rslc and their par files should be '????????.rslc' and '????????.rslc.par'
                               reference:str, # reference date, eg: '20200202'
                               hgt:str, # the DEM in radar coordinate
                               rslc_zarr:str, # output, the flattened rslcs stack in zarr format
                               az_chunk_size:int=-1, # rslcs stack azimuth chunk size, azimuth number of lines by default (one chunk)
                               log:str=None, # logfile, no log by default
                              ):
    '''Generate flatten rslc data from gamma command and convert them into zarr format.
    The shape of hgt should be same as one rslc image, i.e. the hgt file is generated with 1 by 1 look geocoding.
    '''
    logger = get_logger(logfile=log)
    rslcs = _fetch_slc_par_date(rslc_dir)
    with pd.option_context('display.max_colwidth', 0):
        logger.info('rslc found: \n'+str(rslcs))
    rslc_pars = rslcs['par'].to_list()
    dates = rslcs['date'].to_list()
    rslcs = rslcs['rslc'].to_list()

    reference_idx = dates.index(reference)
    ref_rslc = rslcs[reference_idx]
    ref_rslc_par = rslc_pars[reference_idx]
    hgt = Path(hgt)

    n_image = len(rslcs)
    width,nlines = _rdc_width_nlines(ref_rslc_par)
    logger.info(f'number of images: {n_image}.')
    logger.info(f'image number of lines: {nlines}.')
    logger.info(f'image width: {width}.')

    logger.info('run gamma command to generate required data for flattened rslcs:')
    with tempfile.TemporaryDirectory() as temp_dir:
        temp_dir = Path(temp_dir)
        sim_orbs = []
        for i,(date,rslc,rslc_par) in enumerate(zip(dates,rslcs,rslc_pars)):
            off_par = temp_dir/(reference+'_'+date+'.off')
            create_offset_command = f'create_offset {str(ref_rslc_par)} {str(rslc_par)} {str(off_par)} 1 1 1 0 > {str(temp_dir/"log")}'
            logger.info('run command: ' + create_offset_command)
            os.system(create_offset_command)
            # pg.create_offset(ref_rslc_par,rslc_par,off_par,1,1,1,0)
            sim_orb = temp_dir/(reference+'_'+date+'.sim_orb')
            phase_sim_orb_command = f'phase_sim_orb {str(ref_rslc_par)} {str(rslc_par)} {str(off_par)} {str(hgt)} {str(sim_orb)} {str(ref_rslc_par)} - - 1 1 > {str(temp_dir/"log")}'
            logger.info('run command: ' + phase_sim_orb_command)
            os.system(phase_sim_orb_command)
            # pg.phase_sim_orb(ref_rslc_par,rslc_par,off_par,hgt,sim_orb,ref_rslc_par,'-','-',1,1)
            sim_orbs.append(sim_orb)
        logger.info('gamma command finished.')
        logger.info('using dask to load data in gamma binary format to calculate flatten rslcs and save it to zarr.')
        logger.info('starting dask local cluster.')
        cluster = LocalCluster()
        client = Client(cluster)
        logger.info('dask local cluster started.')

        read_gamma_image_delayed = delayed(read_gamma_image, pure=True)
        lazy_rslcs = [read_gamma_image_delayed(rslc,width,dtype='fcomplex') for rslc in rslcs]
        lazy_sim_orbs = [read_gamma_image_delayed(sim_orb,width,dtype='float') for sim_orb in sim_orbs]
        rslcs_data = [da.from_delayed(lazy_rslc,dtype= np.complex64,shape=(nlines,width)) for lazy_rslc in lazy_rslcs]
        rslcs_data = da.stack(rslcs_data,axis=2)
        sim_orb_data =[da.from_delayed(lazy_sim_orb,dtype=np.float32,shape=(nlines,width)) for lazy_sim_orb in lazy_sim_orbs]
        sim_orb_data = da.stack(sim_orb_data,axis=2)
        flatten_rslcs_data = da.exp(sim_orb_data*np.complex64(1j))*rslcs_data
        flatten_rslcs_data = flatten_rslcs_data.rechunk((az_chunk_size,width,1))
        logger.darr_info('flattened rslc', flatten_rslcs_data) 
        _flatten_rslcs_data = flatten_rslcs_data.to_zarr(rslc_zarr,overwrite=True,compute=False)
        logger.info('computing graph setted. doing all the computing.')
        da.compute(_flatten_rslcs_data)
        logger.info('computing finished.')
        cluster.close()
        logger.info('dask cluster closed.')

# %% ../../nbs/CLI/load.ipynb 20
@call_parse
@log_args
def de_load_gamma_lat_lon_hgt(diff_par:str, # geocoding diff_par,using the simulated image as reference
                              rslc_par:str, # par file of the reference rslc
                              dem_par:str, # dem par
                              hgt:str, # DEM in radar coordinate
                              lat_zarr:str, # output, latitude zarr
                              lon_zarr:str, # output, longitude zarr
                              hgt_zarr:str, # output, height zarr
                              az_chunk_size:int=-1, # azimuth chunk size of lat and lon zarr, azimuth number of lines by default (one chunk)
                              log:str=None, # logfile, no log by default
):
    '''
    Function to load longitude and latitude from gamma binary format to zarr.
    '''
    logger = get_logger(logfile=log)
    geo_width = _geo_width_nlines(dem_par)[0]
    rdc_width, rdc_nlines = _rdc_width_nlines(rslc_par)
    logger.info(f'image shape: ({rdc_nlines},{rdc_width})')

    lat_data = zarr.open(lat_zarr,mode='w',shape=(rdc_nlines,rdc_width),chunks = (az_chunk_size,rdc_width), dtype=np.float64)
    lon_data = zarr.open(lon_zarr,mode='w',shape=(rdc_nlines,rdc_width),chunks = (az_chunk_size,rdc_width), dtype=np.float64)
    hgt_data = zarr.open(hgt_zarr,mode='w',shape=(rdc_nlines,rdc_width),chunks = (az_chunk_size,rdc_width), dtype=np.float32)

    with tempfile.TemporaryDirectory() as temp_dir:
        temp_dir = Path(temp_dir)
        logger.info('run gamma command to generate longitude, latitude and height:')
        pt_i = np.arange(rdc_width,dtype=np.int32)
        pt_j = np.arange(rdc_nlines,dtype=np.int32)
        pt_ii,pt_jj = np.meshgrid(pt_i,pt_j)
        pt_ij = np.stack((pt_ii,pt_jj),axis=-1)
        write_gamma_plist(pt_ij,temp_dir/'plist')
        
        # no need to write phgt since it is identical to hgt if all point are selected.
        # hgt_data = read_gamma_image(hgt,width=rdc_width,dtype='float')
        # hgt_data = hgt_data.reshape(-1)
        # pg.write_point_data(hgt_data,temp_dir/'plist',temp_dir/'phgt')
        
        command = f"pt2geo {str(temp_dir/'plist')} - {rslc_par} - {hgt} {dem_par} {diff_par} 1 1 - - {str(temp_dir/'plat_lon')} {str(temp_dir/'phgt_wgs84')} > {str(temp_dir/'log')}"
        logger.info('run command: ' + command)
        os.system(command)
        logger.info('gamma command finished.')
        ptlonlat = read_gamma_plist(str(temp_dir/'plat_lon'),dtype='double')
        lon_data[:], lat_data[:] = ptlonlat[:,0].reshape(rdc_nlines,rdc_width),ptlonlat[:,1].reshape(rdc_nlines,rdc_width)
        hgt_data[:] = read_gamma_image(str(temp_dir/'phgt_wgs84'),width=rdc_width,dtype='float')
        logger.info('Done.')

# %% ../../nbs/CLI/load.ipynb 27
@call_parse
@log_args
def de_load_gamma_look_vector(theta:str, # elevation angle
                              phi:str, # orientation angle
                              lt:str, # lookup table
                              rslc_par:str, # par file of the reference rslc
                              dem_par:str, # dem par
                              theta_zarr:str, # output, elevation angle zarr
                              phi_zarr:str, # output, orientation angle zarr
                              az_chunk_size:int=-1, # azimuth chunk size of lat and lon zarr, azimuth number of lines by default (one chunk)
                              log:str=None, # logfile, no log by default
):
    '''
    Load look vector (elevation angle and orientation angle) in map geometry
    from gamma binary format to look vector in radar geometry zarr file.
    The two input data should be generated with the `look_vector` gamma command.
    '''
    logger = get_logger(logfile=log)
    geo_width = _geo_width_nlines(dem_par)[0]
    rdc_width, rdc_nlines = _rdc_width_nlines(rslc_par)
    logger.info(f'image shape: ({rdc_nlines},{rdc_width})')

    theta_data = zarr.open(theta_zarr,mode='w',shape=(rdc_nlines,rdc_width),chunks = (az_chunk_size,rdc_width), dtype=np.float32)
    phi_data = zarr.open(phi_zarr,mode='w',shape=(rdc_nlines,rdc_width),chunks = (az_chunk_size,rdc_width), dtype=np.float32)

    with tempfile.TemporaryDirectory() as temp_dir:
        temp_dir = Path(temp_dir)
        theta_rdc = temp_dir/'theta_rdc'
        logger.info('run gamma command to generate elevation angle in range doppler coordinate:')
        command = f'geocode {lt} {theta} {geo_width} {str(theta_rdc)} {rdc_width} {rdc_nlines} > {temp_dir/"log"}'
        logger.info('run command: ' + command)
        os.system(command)
        logger.info('gamma command finished.')
        logger.info('writing data.')
        theta_data[:] = read_gamma_image(theta_rdc,rdc_width,dtype='float')
        phi_rdc = temp_dir/'phi_rdc'
        logger.info('run gamma command to generate orientation angle in range doppler coordinate:')
        command = f'geocode {lt} {phi} {geo_width} {str(phi_rdc)} {rdc_width} {rdc_nlines} > {temp_dir/"log"}'
        logger.info('run command: ' + command)
        os.system(command)
        logger.info('gamma command finished.')
        logger.info('writing data.')
        phi_data[:] = read_gamma_image(phi_rdc,rdc_width,dtype='float')
        logger.info('Done.')

# %% ../../nbs/CLI/load.ipynb 33
@call_parse
@log_args
def de_load_gamma_range(rslc_par:str, # par file of one rslc
                        range_zarr:str, # output, range distance zarr
                        az_chunk_size:int=-1, # azimuth chunk size of lat and lon zarr, azimuth number of lines by default (one chunk)
                        log:str=None, # logfile, no log by default
):
    '''
    Generate slant range distance and save to zarr.
    '''
    logger = get_logger(logfile=log)
    rdc_width, rdc_nlines = _rdc_width_nlines(rslc_par)
    logger.info(f'image shape: ({rdc_nlines},{rdc_width})')
    with open(rslc_par) as f:
        for line in f:
            if re.search('near_range_slc',line):
                rho0 = float(line.split()[1])
            if re.search('range_pixel_spacing',line):
                d_rho = float(line.split()[1])

    logger.info('Calculating slant range distance.')
    rho1d = np.arange(rdc_width)*d_rho+rho0
    rho2d = np.tile(rho1d,(rdc_nlines,1))
    range_data = zarr.open(range_zarr,mode='w',shape=(rdc_nlines,rdc_width),chunks = (az_chunk_size,rdc_width), dtype=np.float32)
    logger.info('writing data.')
    range_data[:] = rho2d
    logger.info('Done.')

# %% ../../nbs/CLI/load.ipynb 38
@call_parse
@log_args
def de_load_gamma_metadata(rslc_dir:str, # # gamma rslc directory, the name of the rslc and their par files should be '????????.rslc' and '????????.rslc.par'
                           reference:str, # reference date, eg: '20200202'
                           meta_file:str, # text toml file for meta data
                           log:str=None, # logfile, no log by default
):
    '''
    Load necessary metadata into a toml file.
    '''
    meta = dict()
    rslc_dir = Path(rslc_dir)
    dates = []
    logger = get_logger(logfile=log)
    for par_file in sorted(rslc_dir.glob('*.rslc.par')):
        dates.append(par_file.name[:8])
    meta['dates'] = dates

    with open(rslc_dir/(reference+'.rslc.par')) as f:
        for line in f:
            if re.search('radar_frequency',line):
                radar_f = float(line.split()[1])
                rdr_wavelen = speed_of_light/radar_f
                logger.info('Fetching randar wavelength')
            if re.search('heading',line):
                heading = float(line.split()[1])
                logger.info('Fetching heading angle')
            if re.search('range_pixel_spacing',line):
                dr = float(line.split()[1])
                logger.info('Fetching range pixel spacing')
            if re.search('azimuth_pixel_spacing',line):
                daz = float(line.split()[1])
                logger.info('fetching azimuth pixel spacing')

    meta['radar_wavelength'] = rdr_wavelen
    meta['range_pixel_spacing'] = dr
    meta['azimuth_pixel_spacing'] = daz

    with tempfile.TemporaryDirectory() as temp_dir:        
        temp_dir = Path(temp_dir)
        slc_tab = temp_dir/'slc_tab'
        bperp = temp_dir/'bperp'
        itab = temp_dir/'itab'
        
        tab_content = ''
        for date in dates:
            tab_content += str(rslc_dir/(date+'.rslc'))
            tab_content += '      '
            tab_content += str(rslc_dir/(date+'.rslc.par'))
            tab_content += '\n'
        slc_tab.write_text(tab_content)

        logger.info('Run gamma command to calculate baseline:')
        command = f"base_calc {str(slc_tab)} {str(rslc_dir/(reference+'.rslc.par'))} {str(bperp)} {str(itab)} - > {temp_dir/'log'}"
        logger.info('run command: ' + command)
        os.system(command)
        logger.info('gamma command finished.')
        dat = pd.read_csv(bperp, sep='\s+', header=None)
        base=dat[3].to_numpy()
        base=base.astype(np.float32)

    meta['perpendicular_baseline'] = base

    with open(meta_file,'w') as f:
        a = toml.dump(meta,f,encoder=toml.TomlNumpyEncoder())
    logger.info('All meta data: \n'+a)
    logger.info('writing data in toml file.')
    logger.info('Done.')



# AUTOGENERATED! DO NOT EDIT! File to edit: ../../nbs/CLI/co.ipynb.

# %% auto 0
__all__ = ['de_emperical_co_pc']

# %% ../../nbs/CLI/co.ipynb 3
import math

import zarr
import cupy as cp
import numpy as np
from matplotlib import pyplot as plt
import colorcet

import dask
from dask import array as da
from dask import delayed
from dask.distributed import Client, LocalCluster
from dask_cuda import LocalCUDACluster

from ..co import emperical_co_pc
from .utils.logging import get_logger, log_args
from .utils.dask import pad_internal, get_cuda_cluster
from .utils.chunk_size import get_pc_chunk_size_from_n_az_chunk

from fastcore.script import call_parse

# %% ../../nbs/CLI/co.ipynb 4
@call_parse
@log_args
def de_emperical_co_pc(rslc:str, # input: rslc stack
                       ds_can_is_shp:str, # input: bool array indicating the SHPs of every pixel
                       ds_can_idx:str, # input: bool array indicating DS candidate
                       ds_can_coh:str, # output: complex coherence matrix for DS candidate
                       az_chunk_size:int=None, # azimuth chunk size, optional. Default: the azimuth chunk size in rslc stack
                       n_pc_chunk:int=None, # number of point chunk, optional.
                       pc_chunk_size:int=None, # chunk size of output zarr dataset, optional. Default: same as ds_can_is_shp
                       ds_can_coh_ave_fig:str=None, # path to the plot of average coherence matrix of DS candidate, optional. Default: no plot
                       log=None, # log file. Default: no log file
                       ):
    '''estimate emperical coherence matrix on point cloud data.
    Only one of `n_pc_chunk` and `pc_chunk_size` needs to be setted. The other one is automatically determined.
    If all of them are not setted, the `n_pc_chunk` will be setted as the number of azimuth chunks.
    '''
    rslc_path = rslc
    ds_can_is_shp_path = ds_can_is_shp
    ds_can_idx_path = ds_can_idx
    ds_can_coh_path = ds_can_coh
    logger = get_logger(logfile=log)

    rslc_zarr = zarr.open(rslc_path,mode='r')
    logger.zarr_info(rslc_path, rslc_zarr)
    assert rslc_zarr.ndim == 3, "rslc dimentation is not 3."

    ds_can_is_shp_zarr = zarr.open(ds_can_is_shp_path,mode='r')
    logger.zarr_info(ds_can_is_shp_path, ds_can_is_shp_zarr)
    assert ds_can_is_shp_zarr.ndim == 3, "ds_can_is_shp dimentation is not 3."

    ds_can_idx_zarr = zarr.open(ds_can_idx_path,mode='r')
    logger.zarr_info(ds_can_idx_path, ds_can_idx_zarr)
    assert ds_can_idx_zarr.ndim == 2, "ds_can_idx dimentation is not 2."
    logger.info('loading ds_can_idx into memory.')
    ds_can_idx = zarr.open(ds_can_idx_path,mode='r')[:]

    az_win, r_win = ds_can_is_shp_zarr.shape[1:]
    az_half_win = int((az_win-1)/2)
    r_half_win = int((r_win-1)/2)
    logger.info(f'got azimuth window size and half azimuth window size from is_shp shape: {az_win}, {az_half_win}')
    logger.info(f'got range window size and half range window size from is_shp shape: {r_win}, {r_half_win}')

    if not az_chunk_size:
        az_chunk_size = rslc_zarr.chunks[0]
        logger.info('using default parallel processing azimuth chunk size from rslc dataset.')
    logger.info('parallel processing azimuth chunk size: '+str(az_chunk_size))

    logger.info('starting dask CUDA local cluster.')
    cluster, client = get_cuda_cluster()
    logger.info('dask local CUDA cluster started.')

    logger.info('create raster bool array is_ds_can from ds_can_idx')
    np_is_ds_can = np.zeros(rslc_zarr.shape[:2],dtype=bool)
    np_is_ds_can[(ds_can_idx[0],ds_can_idx[1])] = True
    logger.info('create dask bool array is_ds_can')
    cpu_is_ds_can = da.from_array(np_is_ds_can,chunks=(az_chunk_size,-1))
    logger.darr_info('is_ds_can', cpu_is_ds_can)
    
    logger.info('Using azimuth chunk size as the processing chunk size.')
    logger.info('Calculate point chunk size')
    cpu_ds_can_idx = da.nonzero(cpu_is_ds_can)
    idx_0 = cpu_ds_can_idx[0]
    process_pc_chunk_size = idx_0.compute_chunk_sizes().chunks[0]
    logger.info(f'Point chunk size: {process_pc_chunk_size}')

    cpu_rslc = da.from_zarr(rslc_path,chunks=(az_chunk_size,*rslc_zarr.shape[1:]))
    logger.darr_info('rslc', cpu_rslc)

    cpu_ds_can_is_shp = da.from_zarr(ds_can_is_shp_path,chunks=(process_pc_chunk_size,*ds_can_is_shp_zarr.shape[1:]))
    logger.darr_info('ds_can_is_shp', cpu_ds_can_is_shp)

    depth = {0:az_half_win, 1:r_half_win, 2:0}; boundary = {0:'none',1:'none',2:'none'}
    cpu_rslc_overlap = da.overlap.overlap(cpu_rslc,depth=depth, boundary=boundary)
    logger.info('setting shared boundaries between rlsc chunks.')
    logger.darr_info('rslc_overlap', cpu_rslc_overlap)

    depth = {0:az_half_win, 1:r_half_win}
    cpu_is_ds_can_padded = pad_internal(cpu_is_ds_can,depth=depth)
    logger.info('padding zero between is_ds_can chunks.')
    logger.darr_info('is_ds_can_padded', cpu_is_ds_can_padded)

    logger.info(f'estimating coherence matrix.')
    rslc_overlap = cpu_rslc_overlap.map_blocks(cp.asarray)
    is_ds_can_padded = cpu_is_ds_can_padded.map_blocks(cp.asarray)
    ds_can_is_shp = cpu_ds_can_is_shp.map_blocks(cp.asarray)
    
    emperical_co_pc_delayed = delayed(emperical_co_pc,pure=True,nout=2)

    is_ds_can_padded_delayed = is_ds_can_padded.to_delayed()
    is_ds_can_padded_delayed = np.squeeze(is_ds_can_padded_delayed,axis=-1)
    rslc_overlap_delayed = rslc_overlap.to_delayed()
    rslc_overlap_delayed = np.squeeze(rslc_overlap_delayed,axis=(-2,-1))
    ds_can_is_shp_delayed = ds_can_is_shp.to_delayed()
    ds_can_is_shp_delayed = np.squeeze(ds_can_is_shp_delayed,axis=(-2,-1))

    ds_can_coh_delayed = np.empty_like(rslc_overlap_delayed,dtype=object)
    ds_can_idx_delayed = np.empty_like(rslc_overlap_delayed,dtype=object)

    nimage = rslc_overlap.shape[-1]
    with np.nditer(rslc_overlap_delayed,flags=['multi_index','refs_ok'], op_flags=['readwrite']) as it:
        for block in it:
            idx = it.multi_index
            ds_can_idx_delayed[idx] = delayed(cp.where)(is_ds_can_padded_delayed[idx])
            ds_can_coh_delayed[idx] = emperical_co_pc_delayed(rslc_overlap_delayed[idx],ds_can_idx_delayed[idx],ds_can_is_shp_delayed[idx])[1]
            chunk_shape = (ds_can_is_shp.blocks[idx].shape[0],nimage,nimage)
            dtype = rslc_overlap.dtype
            ds_can_coh_delayed[idx] = da.from_delayed(ds_can_coh_delayed[idx],shape=chunk_shape,meta=cp.array((),dtype=dtype))

    ds_can_coh = da.block(ds_can_coh_delayed.reshape(*ds_can_coh_delayed.shape,1,1).tolist())
    cpu_ds_can_coh = ds_can_coh.map_blocks(cp.asnumpy)
    logger.info(f'got coherence matrix.')

    # zarr do not support irregular chunk size
    pc_chunk_size = get_pc_chunk_size_from_n_az_chunk('rslc','ds_can_coh',cpu_rslc.shape[0],az_chunk_size,ds_can_coh.shape[0],logger,pc_chunk_size=pc_chunk_size,n_pc_chunk=n_pc_chunk)
    cpu_ds_can_coh = cpu_ds_can_coh.rechunk((pc_chunk_size,nimage,nimage))
    logger.info('rechunking ds_can_coh to chunk size (for saving with zarr): '+str(cpu_ds_can_coh.chunksize))

    logger.info('saving ds_can_coh.')
    _cpu_ds_can_coh = cpu_ds_can_coh.to_zarr(ds_can_coh_path,overwrite=True,compute=False)

    cpu_ds_can_coh_ave = da.abs(cpu_ds_can_coh).mean(axis=0)

    logger.info('computing graph setted. doing all the computing.')
    #This function is really slow just because the coherence is very big and rechunk and saving takes too much time.
    # I do not find any solution to it.
    cpu_ds_can_coh_ave_result = da.compute(_cpu_ds_can_coh,cpu_ds_can_coh_ave)[1]

    # pdb.set_trace()
    logger.info('computing finished.')
    cluster.close()
    logger.info('dask cluster closed.')

    if ds_can_coh_ave_fig:
        logger.info('plotting average coherence matrix.')
        fig, ax = plt.subplots(1,1,figsize=(15,10))
        pcm = ax.imshow(cpu_ds_can_coh_ave_result,cmap=colorcet.cm.fire)
        ax.set(title='Average Coherence Matrix',xlabel='Image Index',ylabel='Image Index')
        fig.colorbar(pcm)
        fig.show()
        fig.savefig(ds_can_coh_ave_fig)



# AUTOGENERATED! DO NOT EDIT! File to edit: ../../nbs/CLI/pl.ipynb.

# %% auto 0
__all__ = ['de_emi']

# %% ../../nbs/CLI/pl.ipynb 4
import math

import zarr
import cupy as cp
import numpy as np
from matplotlib import pyplot as plt
import colorcet

import dask
from dask import array as da
from dask import delayed
from dask.distributed import Client, LocalCluster
from dask_cuda import LocalCUDACluster

from ..pl import emi
from ..pc import pc2ras
from .utils.logging import get_logger, log_args
from .utils.dask import get_cuda_cluster

from fastcore.script import call_parse

# %% ../../nbs/CLI/pl.ipynb 5
@call_parse
@log_args
def de_emi(coh:str, # coherence matrix
           ph:str, # output, wrapped phase
           emi_quality:str, #output, pixel quality
           ref:int=0, # reference image for phase
           point_chunk_size:int=None, # parallel processing point chunk size
           log:str=None, #log
           plot_emi_quality:bool=False, # if plot the emi quality
           vmin:int=1.0, # min value of emi quality to plot
           vmax:int=1.3, # max value of emi quality to plot
           ds_idx:str=None, # index of ds
           shape:tuple=None, # shape of one image
           emi_quality_fig:str=None, # path to save the emi quality plot, optional. Default, no saving
          ):
    coh_path = coh
    ph_path = ph
    emi_quality_path = emi_quality
    ds_idx_path = ds_idx

    logger = get_logger(logfile=log)
    coh_zarr = zarr.open(coh_path,mode='r')
    logger.zarr_info(coh_path,coh_zarr)

    if not point_chunk_size:
        point_chunk_size = coh_zarr.chunks[0]
        logger.info('using default parallel processing azimuth chunk size from coh dataset.')
    logger.info('parallel processing point chunk size: '+str(point_chunk_size))

    logger.info('starting dask CUDA local cluster.')
    cluster = LocalCUDACluster()
    client = Client(cluster)
    # for no reason, here when threads_per_worker is not 1, there will be some error.
    logger.info('dask local CUDA cluster started.')

    cpu_coh = da.from_zarr(coh_path, chunks=(point_chunk_size,*coh_zarr.shape[1:]))
    logger.darr_info('coh', cpu_coh)

    logger.info(f'phase linking with EMI.')
    coh = cpu_coh.map_blocks(cp.asarray)
    coh_delayed = coh.to_delayed()
    coh_delayed = np.squeeze(coh_delayed,axis=(-2,-1))

    ph_delayed = np.empty_like(coh_delayed,dtype=object)
    emi_quality_delayed = np.empty_like(coh_delayed,dtype=object)

    with np.nditer(coh_delayed,flags=['multi_index','refs_ok'], op_flags=['readwrite']) as it:
        for block in it:
            idx = it.multi_index
            ph_delayed[idx], emi_quality_delayed[idx] = delayed(emi,pure=True,nout=2)(coh_delayed[idx])
            ph_delayed[idx] = da.from_delayed(ph_delayed[idx],shape=coh.blocks[idx].shape[0:2],meta=cp.array((),dtype=coh.dtype))
            emi_quality_delayed[idx] = da.from_delayed(emi_quality_delayed[idx],shape=coh.blocks[idx].shape[0:1],meta=cp.array((),dtype=cp.float32))

    ph = da.block(ph_delayed[...,None].tolist())
    emi_quality = da.block(emi_quality_delayed.tolist())

    cpu_ph = ph.map_blocks(cp.asnumpy)
    cpu_emi_quality = emi_quality.map_blocks(cp.asnumpy)
    logger.info(f'got ph and emi_quality.')
    logger.darr_info('ph', cpu_ph)
    logger.darr_info('emi_quality', cpu_emi_quality)

    logger.info('saving ph and emi_quality.')
    _cpu_ph = cpu_ph.to_zarr(ph_path,compute=False,overwrite=True)
    _cpu_emi_quality = cpu_emi_quality.to_zarr(emi_quality_path,compute=False,overwrite=True)
    
    if not plot_emi_quality:
        logger.info('computing graph setted. doing all the computing.')
        da.compute(_cpu_ph,_cpu_emi_quality)
        logger.info('computing finished.')
        cluster.close()
        logger.info('dask cluster closed.')
    else:
        logger.info('computing graph setted. doing all the computing.')
        emi_quality_result = da.compute(cpu_emi_quality,_cpu_ph,_cpu_emi_quality)[0]
        logger.info('computing finished.')
        cluster.close()
        logger.info('dask cluster closed.')

        logger.info('plotting emi_quality.')
        logger.info('reading is_ds bool array')
        ds_idx_result = zarr.open(ds_idx_path,mode='r')[:]

        emi_quality_raster = pc2ras(ds_idx_result,emi_quality_result,shape)
        logger.info('converting emi_quality from point cloud to raster.')
        fig, ax = plt.subplots(1,1)
        pcm = ax.imshow(emi_quality_raster,interpolation='nearest',vmin=vmin,vmax=vmax)
        ax.set(title='EMI quality factor',xlabel='Range Index',ylabel='Azimuth Index')
        fig.colorbar(pcm)
        if emi_quality_fig:
            logger.info('saving figure')
            fig.savefig(emi_quality_fig)
        logger.info('showing')
        fig.show()



# AUTOGENERATED! DO NOT EDIT! File to edit: ../../../nbs/CLI/utils/chunk_size.ipynb.

# %% auto 0
__all__ = ['get_pc_chunk_size_from_n_pc_chunk', 'get_pc_chunk_size_from_pc_chunk_size', 'get_pc_chunk_size_from_n_az_chunk',
           'get_az_chunk_size_from_n_pc_chunk', 'get_az_chunk_size_from_az_chunk_size']

# %% ../../../nbs/CLI/utils/chunk_size.ipynb 3
import math

# %% ../../../nbs/CLI/utils/chunk_size.ipynb 4
def get_pc_chunk_size_from_n_pc_chunk(in_pc_name, out_pc_name,
                                      n_pc_in, pc_chunk_size_in,
                                      n_pc,
                                      logger,
                                      pc_chunk_size=None,n_pc_chunk=None):
    if pc_chunk_size is not None:
        logger.info(f'got pc_chunk_size for {out_pc_name}: {pc_chunk_size}')
        return pc_chunk_size
    else:
        if n_pc_chunk is not None:
            logger.info(f'got n_pc_chunk for {out_pc_name}: {n_pc_chunk}')
            logger.info(f'automatically determine pc_chunk_size from n_pc of {out_pc_name} and n_pc_chunk of {out_pc_name}')
            pc_chunk_size = math.ceil(n_pc/n_pc_chunk)
            logger.info(f'pc_chunk_size for {out_pc_name}: {pc_chunk_size}')
            return pc_chunk_size
        else:
            logger.info(f'automatically determine pc_chunk_size from n_pc of {out_pc_name} and n_pc_chunk of {in_pc_name}')
            n_pc_chunk = math.ceil(n_pc_in/pc_chunk_size_in)
            pc_chunk_size = math.ceil(n_pc/n_pc_chunk)
            logger.info(f'pc_chunk_size for {out_pc_name}: {pc_chunk_size}')
            return pc_chunk_size

# %% ../../../nbs/CLI/utils/chunk_size.ipynb 5
def get_pc_chunk_size_from_pc_chunk_size(in_pc_name, out_pc_name,
                                         pc_chunk_size_in,
                                         n_pc,
                                         logger,
                                         pc_chunk_size=None,n_pc_chunk=None):
    if pc_chunk_size is not None:
        logger.info(f'got pc_chunk_size for {out_pc_name}: {pc_chunk_size}')
        return pc_chunk_size
    else:
        if n_pc_chunk is not None:
            logger.info(f'got n_pc_chunk for {out_pc_name}: {n_pc_chunk}')
            logger.info(f'automatically determine pc_chunk_size from n_pc of {out_pc_name} and n_pc_chunk of {out_pc_name}')
            pc_chunk_size = math.ceil(n_pc/n_pc_chunk)
            logger.info(f'pc_chunk_size for {out_pc_name}: {pc_chunk_size}')
            return pc_chunk_size
        else:
            logger.info(f'automatically determine pc_chunk_size from pc_chunk_size of {in_pc_name}.')
            pc_chunk_size = pc_chunk_size_in
            logger.info(f'pc_chunk_size for {out_pc_name}: {pc_chunk_size}')
            return pc_chunk_size

# %% ../../../nbs/CLI/utils/chunk_size.ipynb 6
def get_pc_chunk_size_from_n_az_chunk(ras_name, pc_name,
                                      nlines, az_chunk_size,
                                      n_pc,
                                      logger,
                                      pc_chunk_size=None,n_pc_chunk=None):
    if pc_chunk_size is not None:
        logger.info(f'got pc_chunk_size for {pc_name}: {pc_chunk_size}')
        return pc_chunk_size
    else:
        if n_pc_chunk is not None:
            logger.info(f'got n_pc_chunk for {pc_name}: {n_pc_chunk}')
            logger.info(f'automatically determine pc_chunk_size from n_pc of {pc_name} and n_pc_chunk of {pc_name}')
            pc_chunk_size = math.ceil(n_pc/n_pc_chunk)
            logger.info(f'pc_chunk_size for {pc_name}: {pc_chunk_size}')
            return pc_chunk_size
        else:
            logger.info(f'automatically determine pc_chunk_size from n_pc of {pc_name} and n_az_chunk of {ras_name}')
            n_pc_chunk = math.ceil(nlines/az_chunk_size)
            pc_chunk_size = math.ceil(n_pc/n_pc_chunk)
            logger.info(f'pc_chunk_size for {pc_name}: {pc_chunk_size}')
            return pc_chunk_size

# %% ../../../nbs/CLI/utils/chunk_size.ipynb 7
def get_az_chunk_size_from_n_pc_chunk(pc_name, ras_name,
                                      n_pc, pc_chunk_size,
                                      nlines,
                                      logger,
                                      az_chunk_size=None,n_az_chunk=None):
    if az_chunk_size is not None:
        logger.info(f'got az_chunk_size for {ras_name}: {az_chunk_size}')
        return az_chunk_size
    else:
        if n_az_chunk is not None:
            logger.info(f'got n_az_chunk for {ras_name}: {n_az_chunk}')
            logger.info(f'automatically determine az_chunk_size from nlines of {ras_name} and n_pc_chunk of {ras_name}')
            az_chunk_size = math.ceil(nlines/n_az_chunk)
            logger.info(f'az_chunk_size for {ras_name}: {az_chunk_size}')
            return az_chunk_size
        else:
            logger.info(f'automatically determine az_chunk_size from nlines of {ras_name} and n_pc_chunk of {pc_name}')
            n_az_chunk = math.ceil(n_pc/pc_chunk_size)
            az_chunk_size = math.ceil(nlines/n_az_chunk)
            logger.info(f'az_chunk_size for {az_name}: {az_chunk_size}')
            return az_chunk_size

# %% ../../../nbs/CLI/utils/chunk_size.ipynb 8
def get_az_chunk_size_from_az_chunk_size(in_ras_name, out_ras_name,
                                         az_chunk_size_in,
                                         nlines,
                                         logger,
                                         az_chunk_size=None,n_az_chunk=None):
    if az_chunk_size is not None:
        logger.info(f'got az_chunk_size for {out_ras_name}: {az_chunk_size}')
        return az_chunk_size
    else:
        if n_az_chunk is not None:
            logger.info(f'got n_az_chunk for {out_ras_name}: {n_az_chunk}')
            logger.info(f'automatically determine az_chunk_size from nlines of {out_ras_name} and n_pc_chunk of {out_ras_name}')
            az_chunk_size = math.ceil(nlines/n_az_chunk)
            logger.info(f'az_chunk_size for {out_ras_name}: {az_chunk_size}')
            return az_chunk_size
        else:
            logger.info(f'automatically determine az_chunk_size from az_chunk_size of {in_ras_name}')
            az_chunk_size = az_chunk_size_in
            logger.info(f'az_chunk_size for {out_az_name}: {az_chunk_size}')
            return az_chunk_size



# AUTOGENERATED! DO NOT EDIT! File to edit: ../nbs/API/shp.ipynb.

# %% auto 0
__all__ = ['ks_test']

# %% ../nbs/API/shp.ipynb 4
import cupy as cp
import math

# %% ../nbs/API/shp.ipynb 6
# It looks cupy do not support pointer to pointer: float** rmli_stack
# These code are modified from https://dev.thep.lu.se/yat under GPLv3 license.
_ks_test_kernel = cp.ElementwiseKernel(
    'raw T rmli_stack, int32 nlines, int32 width, int32 nimages, int32 az_half_win, int32 r_half_win',
    'raw T dist, raw T p',
    '''
    int az_win = 2*az_half_win+1;
    int r_win = 2*r_half_win+1;
    int win = az_win*r_win;
    
    int ref_idx = i/win;
    int ref_az = ref_idx/width;
    int ref_r = ref_idx -ref_az*width;
    
    int win_idx = i - ref_idx*win;
    int win_az = win_idx/r_win;
    int win_r = win_idx - win_az*r_win;
    int sec_az = ref_az + win_az - az_half_win;
    int sec_r = ref_r + win_r - r_half_win;
    int sec_idx = sec_az*width + sec_r;
    
    if (ref_r >= width && ref_az >= nlines) {
        return;
    }
    if (sec_az < 0 || sec_az >= nlines || sec_r < 0 || sec_r >= width) {
        dist[ref_idx*win+win_az*r_win+win_r] = -1.0;
        p[ref_idx*win+win_az*r_win+win_r] = -1.0;
        return;
    }
    
    // Compute the maximum difference between the cumulative distributions
    int j1 = 0, j2 = 0;
    T f1, f2, d, dmax = 0.0, en = nimages;

    while (j1 < nimages && j2 < nimages) {
        f1 = rmli_stack[ref_idx*nimages + j1];
        f2 = rmli_stack[sec_idx*nimages + j2];
        if (f1 <= f2) j1++;
        if (f1 >= f2) j2++;
        d = fabs((j2-j1)/en);
        if (d > dmax) dmax = d;
    }
    en=sqrt(en/2);
    p[ref_idx*win+win_az*r_win+win_r] = ks_p((en+0.12+0.11/en)*dmax);
    dist[ref_idx*win+win_az*r_win+win_r] = dmax;
    ''',
    name = 'ks_test_kernel',no_return=True,
    preamble = '''
    __device__ T ks_p(T x)
    {
        T x2 = -2.0*x*x;
        int sign = 1;
        T p = 0.0,p2 = 0.0;
    
        for (int i = 1; i <= 100; i++) {
            p += sign*2*exp(x2*i*i);
            if (p==p2) return p;
            sign = -sign;
            p2 = p;
        }
        return p;
    }
    ''',)

# %% ../nbs/API/shp.ipynb 7
def ks_test(rmli:cp.ndarray, # the rmli stack, dtype: cupy.floating
            az_half_win:int, # SHP identification half search window size in azimuth direction
            r_half_win:int, # SHP identification half search window size in range direction
            block_size:int=128, # the CUDA block size, it only affects the calculation speed
           ) -> tuple[cp.ndarray,cp.ndarray] : # the KS test statistics `dist` and p value `p`
    '''
    SHP identification based on Two-Sample Kolmogorov-Smirnov Test.
    '''
    sorted_rmli = cp.sort(rmli,axis=-1) # In fact, this step is most time consuming, consider move it out
    az_win = 2*az_half_win+1
    r_win = 2*r_half_win+1
    nlines = rmli.shape[0]
    width = rmli.shape[1]
    nimages = rmli.shape[-1]
    dist = cp.empty((nlines,width,az_win,r_win),dtype=rmli.dtype)
    p = cp.empty((nlines,width,az_win,r_win),dtype=rmli.dtype)

    _ks_test_kernel(sorted_rmli,cp.int32(nlines),cp.int32(width),cp.int32(nimages),
                    cp.int32(az_half_win),cp.int32(r_half_win),dist,p,
                    size=width*nlines*r_win*az_win,block_size=block_size)
    return dist,p



# AUTOGENERATED! DO NOT EDIT! File to edit: ../../../nbs/CLI/utils/dask.ipynb.

# %% auto 0
__all__ = ['pad_internal', 'get_cuda_cluster', 'get_pc_chunk_size']

# %% ../../../nbs/CLI/utils/dask.ipynb 3
from dask import array as da
import numpy as np
import os
import math
from .appconfig import get_config

# %% ../../../nbs/CLI/utils/dask.ipynb 4
def _pad_chunks(original_chunks, axes):
    """Get new chunks for array with block padding."""
    # this function is adapt from https://github.com/dask/dask/blob/main/dask/array/overlap.py#L17
    chunks = []
    for i, bds in enumerate(original_chunks):
        depth = axes.get(i, 0)
        if isinstance(depth, tuple):
            left_depth = depth[0]
            right_depth = depth[1]
        else:
            left_depth = depth
            right_depth = depth
        
        chunk = []
        for bd in bds:
            chunk.append(left_depth+bd+right_depth)
        chunks.append(chunk)
    return chunks

# %% ../../../nbs/CLI/utils/dask.ipynb 5
def pad_internal(arr:da.array,
                 depth:dict=None):
    '''Pad zero between block boundaries, currently one pad zero are supported'''
    chunks = _pad_chunks(arr.chunks,axes=depth)

    depth_len = len(depth)
    pad_arg = [None]*depth_len
    slice_arg = [None]*depth_len
    for i in range(depth_len):
        if isinstance(depth[i],tuple):
            pad_arg[i] = depth[i]
            left_slice = depth[i][0]
            if depth[i][1] == 0:
                right_slice = None
            else:
                right_slice = -depth[i][-1]
            slice_arg[i] = slice(left_slice,right_slice)
        else:
            pad_arg[i] = (depth[i],depth[i])
            left_slice = depth[i]
            if depth[i] == 0:
                right_slice = None
            else:
                right_slice = -depth[i]
            slice_arg[i] = slice(left_slice,right_slice)
    pad_arg = tuple(pad_arg)
    slice_arg = tuple(slice_arg)
    arr_out = da.map_blocks(np.pad,arr,pad_arg,chunks=chunks)

    return arr_out[slice_arg]

# %% ../../../nbs/CLI/utils/dask.ipynb 6
def get_cuda_cluster():
    from dask.distributed import Client
    from dask_cuda import LocalCUDACluster
    config = get_config()
    
    visible_gpus = config.get('visible_gpus')
    if visible_gpus is None:
        n_gpu = len(os.environ['CUDA_VISIBLE_DEVICES'].split(','))
    else:
        n_gpu = len(visible_gpus)
    n_cpu = config.get('n_cpu')
    threads_per_worker = math.floor(n_cpu/n_gpu)
    cluster = LocalCUDACluster(CUDA_VISIBLE_DEVICES=config.get('visible_gpus'),threads_per_worker=threads_per_worker)
    client = Client(cluster)
    return cluster, client

# %% ../../../nbs/CLI/utils/dask.ipynb 7
def get_pc_chunk_size(az_chunk_size,pc_size,pc_chunk_size=None,n_pc_chunk=None,logger=None):
    '''automatically determine pc_chunk_size, n_pc_chunk.
    '''
    if pc_chunk_size is None:
        if n_pc_chunk is None:
            pc_chunk_size = az_chunk_size
            logger.info(f'set pc_chunk_size as az_chunk_size: {az_chunk_size}')
        else:
            logger.info(f'got n_pc_chunk: {n_pc_chunk}')
            pc_chunk_size = math.ceil(pc_size/pt_n_chunk)
            logger.info(f'set pc_chunk_size as: {pc_chunk_size}')
    else:
        logger.info(f'got pc_chunk_size: {pc_chunk_size}')
    return pc_chunk_size



# AUTOGENERATED! DO NOT EDIT! File to edit: ../../nbs/CLI/shp.ipynb.

# %% auto 0
__all__ = ['de_shp_test', 'de_select_ds_can']

# %% ../../nbs/CLI/shp.ipynb 3
from itertools import product
import math

import zarr
import cupy as cp
import numpy as np
from matplotlib import pyplot as plt
import colorcet

import dask
from dask import array as da
from dask import delayed
from dask.distributed import Client, LocalCluster
from dask_cuda import LocalCUDACluster

from ..shp import ks_test
from .utils.logging import get_logger, log_args
from .utils.dask import get_cuda_cluster
from .utils.chunk_size import get_pc_chunk_size_from_n_az_chunk
from fastcore.script import call_parse

# %% ../../nbs/CLI/shp.ipynb 4
@call_parse
@log_args
def de_shp_test(rslc:str, # input: rslc stack
                pvalue:str, # output: the p value of the test
                az_half_win:int, # azimuth half window size
                r_half_win:int, # range half window size
                method:str=None, # SHP identification method,optional. Default: ks
                az_chunk_size:int=None, # azimuth chunk size, optional. Default: the azimuth chunk size in rslc
                log:str=None, # log file, optional. Default: no log file
               ):
    '''SHP identification through hypothetic test.'''
    rslc_path = rslc
    pvalue_path = pvalue

    logger = get_logger(logfile=log,level='debug')
    if not method: method = 'ks'
    logger.info(f'hypothetic test method: {method}')
    if method != 'ks':
        logger.warning('Currently only KS test is implented. Switching to it.')
        method = 'ks'

    rslc_zarr = zarr.open(rslc_path,mode='r')
    logger.zarr_info(rslc_path,rslc_zarr)

    assert rslc_zarr.ndim == 3, " rslcs dimentation is not 3."

    if not az_chunk_size:
        az_chunk_size = rslc_zarr.chunks[0]
        logger.info('using default parallel processing azimuth chunk size from input rslc.')
    logger.info('parallel processing azimuth chunk size: '+str(az_chunk_size))

    chunks=(az_chunk_size,*rslc_zarr.shape[1:])
    
    logger.info('starting dask CUDA local cluster.')
    cluster, client = get_cuda_cluster()
    # client = Client(cluster)
    logger.info('dask local CUDA cluster started.')

    cpu_rslc = da.from_zarr(rslc_path,chunks=chunks)
    logger.darr_info('rslc',cpu_rslc)

    az_win = 2*az_half_win+1
    logger.info(f'azimuth half window size: {az_half_win}; azimuth window size: {az_win}')
    r_win = 2*r_half_win+1
    logger.info(f'range half window size: {r_half_win}; range window size: {r_win}')

    depth = {0:az_half_win, 1:r_half_win, 2:0}; boundary = {0:'none',1:'none',2:'none'}
    cpu_rslc_overlap = da.overlap.overlap(cpu_rslc,depth=depth, boundary=boundary)
    logger.info('setting shared boundaries between rlsc chunks.')
    logger.darr_info('rslc with overlap', cpu_rslc_overlap)

    rslc_overlap = cpu_rslc_overlap.map_blocks(cp.asarray)
    rmli_overlap = da.abs(rslc_overlap)**2
    logger.darr_info('rmli with overlap', rmli_overlap)

    sorted_rmli_overlap = rmli_overlap.map_blocks(cp.sort,axis=-1)

    delayed_ks_test = delayed(ks_test,pure=True,nout=2)
    rmli_delayed = sorted_rmli_overlap.to_delayed()
    p_delayed = np.empty_like(rmli_delayed,dtype=object)
    dist_delayed = np.empty_like(rmli_delayed,dtype=object)

    logger.info('applying test on sorted rmli stack.')
    with np.nditer(p_delayed,flags=['multi_index','refs_ok'], op_flags=['readwrite']) as p_it:
        for p_block in p_it:
            idx = p_it.multi_index
            dist_delayed[idx],p_delayed[idx] = delayed_ks_test(rmli_delayed[idx],az_half_win=az_half_win,r_half_win=r_half_win)

            chunk_shape = (*sorted_rmli_overlap.blocks[idx].shape[:-1],az_win,r_win)
            dtype = sorted_rmli_overlap.dtype
            # dist_delayed[idx] = da.from_delayed(dist_delayed[idx],shape=chunk_shape,meta=cp.array((),dtype=dtype))
            p_delayed[idx] = da.from_delayed(p_delayed[idx],shape=chunk_shape,meta=cp.array((),dtype=dtype))
    
    p = da.block(p_delayed.reshape(*p_delayed.shape,1).tolist())
    # dist = da.block(dist_delayed.reshape(*dist_delayed.shape,1).tolist())
    logger.info('p value generated')
    logger.darr_info('p value', p)

    depth = {0:az_half_win, 1:r_half_win, 2:0, 3:0}; boundary = {0:'none',1:'none',2:'none',3:'none'}
    # dist = da.overlap.trim_overlap(dist,depth=depth,boundary=boundary)
    p = da.overlap.trim_overlap(p,depth=depth,boundary=boundary)
    logger.info('trim shared boundaries between p value chunks')
    logger.darr_info('trimmed p value', p)

    # cpu_dist = da.map_blocks(cp.asnumpy,dist)
    cpu_p = da.map_blocks(cp.asnumpy,p)
    # cpu_p = cpu_p.rechunk((*cpu_p.chunksize[:2],1,1))
    
    # _cpu_dist = cpu_dist.to_zarr(statistic,overwrite=True,compute=False)
    _cpu_p = cpu_p.to_zarr(pvalue_path,overwrite=True,compute=False)
    logger.info('saving p value.')
    logger.info('computing graph setted. doing all the computing.')
    da.compute(_cpu_p)
    logger.info('computing finished.')
    # pdb.set_trace()
    cluster.close()
    logger.info('dask cluster closed.')

# %% ../../nbs/CLI/shp.ipynb 10
@call_parse
@log_args
def de_select_ds_can(pvalue:str, # input: pvalue of hypothetic test
                     ds_can_idx:str, # output: index array of DS candidate
                     ds_can_is_shp:str, # output: bool array indicating the SHPs of DS candidate
                     p_max:float=0.05, # threshold of p value to select SHP,optional. Default: 0.05
                     shp_num_min:int=50, # threshold of number of SHPs to select DS candidate,optional. Default: 50
                     az_chunk_size:int=None, # azimuth chunk size, optional. Default: the azimuth chunk size in pvalue
                     n_pc_chunk:int=None, # number of point target chunks, optional.
                     pc_chunk_size:int=None, # point target chunk size, optional. 
                     shp_num_fig:str=None, # path to the plot of number of SHPs, optional. Default: no plot
                     is_ds_can_fig:str=None, # path to the plot of DSs candidate distribution, optional. Default: no plot
                     log=None, # log file. Default: no log file
                     ):
    '''
    Select DS candidate based on pvalue of SHP test.
    Only one of `n_pc_chunk` and `pc_chunk_size` needs to be setted. The other one is automatically determined.
    If all of them are not setted, the `n_pc_chunk` will be setted as the number of azimuth chunks.
    '''

    logger = get_logger(logfile=log)

    p_zarr = zarr.open(pvalue,mode='r')
    logger.zarr_info(pvalue, p_zarr)

    assert p_zarr.ndim == 4, " pvalue dimentation is not 4."

    if not az_chunk_size:
        az_chunk_size = p_zarr.chunks[0]
        logger.info('using default parallel processing azimuth chunk size from input pvalue zarr.')
    logger.info('parallel processing azimuth chunk size: '+str(az_chunk_size))

    chunks=(az_chunk_size,*p_zarr.shape[1:])

    logger.info('starting dask local cluster.')
    cluster = LocalCluster()
    client = Client(cluster)
    logger.info('dask local cluster started.')

    p = da.from_zarr(pvalue,chunks=chunks)
    logger.darr_info('pvalue', p)

    ds_can_is_shp_path= ds_can_is_shp
    ds_can_idx_path = ds_can_idx

    is_shp = (p < p_max) & (p >= 0)
    logger.info('selecting SHPs based on pvalue threshold: '+str(p_max))
    logger.darr_info('is_shp', is_shp)

    shp_num = da.count_nonzero(is_shp,axis=(-2,-1))
    is_ds_can = shp_num >= shp_num_min
    logger.info('selecting DS candidates based on minimum of number of SHPs: '+str(shp_num_min))
    logger.darr_info('is_ds_can', is_ds_can)
    
    logger.info('calculate ds_can index:')
    ds_can_idx = da.nonzero(is_ds_can)
    ds_can_idx = da.stack(ds_can_idx,allow_unknown_chunksizes=True)
    logger.info('slice is_shp on ds_can index:')
    with dask.config.set(**{'array.slicing.split_large_chunks': False}):
        ds_can_is_shp = is_shp.reshape(-1,*is_shp.shape[2:])[is_ds_can.reshape(-1)]
    # ds_can_is_shp, ds_can_idx, is_ds_can, shp_num = client.persist((ds_can_is_shp,ds_can_idx,is_ds_can,shp_num))

    ds_can_is_shp.compute_chunk_sizes()
    ds_can_idx.compute_chunk_sizes()
    
    logger.darr_info('ds_can_idx',ds_can_idx)
    logger.darr_info('ds_can_is_shp',ds_can_is_shp)
    
    pc_chunk_size = get_pc_chunk_size_from_n_az_chunk('p_value','ds_can_idx', p.shape[0], az_chunk_size, ds_can_idx.shape[1], logger, n_pc_chunk=n_pc_chunk,pc_chunk_size=pc_chunk_size)

    logger.info(f'rechunk ds_can_idx and ds_can_is_shp:')
    ds_can_is_shp = ds_can_is_shp.rechunk((pc_chunk_size,*ds_can_is_shp.shape[1:]))
    ds_can_idx = ds_can_idx.rechunk((2,pc_chunk_size))
    
    logger.darr_info('ds_can_idx',ds_can_idx)
    logger.darr_info('ds_can_is_shp',ds_can_is_shp)

    _ds_can_is_shp = ds_can_is_shp.to_zarr(ds_can_is_shp_path,overwrite=True,compute=False)
    logger.info('saving ds_can_is_shp.')
    _ds_can_idx = ds_can_idx.to_zarr(ds_can_idx_path,overwrite=True,compute=False)
    logger.info('saving ds_can_idx.')

    logger.info('computing graph setted. doing all the computing.')
    shp_num_result, is_ds_can_result = da.compute(_ds_can_is_shp,_ds_can_idx,shp_num,is_ds_can)[2:]
    logger.info('computing finished.')
    cluster.close()
    logger.info('dask cluster closed.')

    if shp_num_fig:
        logger.info('plotting number of SHPs.')
        fig, ax = plt.subplots(1,1,figsize=(10,10))
        pcm = ax.imshow(shp_num_result,cmap=colorcet.cm.fire)
        ax.set(title='Number of SHPs',xlabel='Range Index',ylabel='Azimuth Index')
        fig.colorbar(pcm)
        fig.savefig(shp_num_fig)
        fig.show()
    
    if is_ds_can_fig:
        logger.info('plotting DS candidate distribution.')
        fig, ax = plt.subplots(1,1,figsize=(10,10))
        pcm = ax.imshow(is_ds_can_result,cmap=colorcet.cm.fire)
        ax.set(title='DS Candidate distribution',xlabel='Range Index',ylabel='Azimuth Index')
        fig.colorbar(pcm)
        fig.savefig(is_ds_can_fig)
        fig.show()



# AUTOGENERATED! DO NOT EDIT! File to edit: ../nbs/API/pc.ipynb.

# %% auto 0
__all__ = ['pc2ras', 'pc_union', 'pc_intersect', 'pc_diff']

# %% ../nbs/API/pc.ipynb 4
import cupy as cp
import numpy as np
from typing import Union
from cupy._sorting.search import _exists_kernel

# %% ../nbs/API/pc.ipynb 5
def pc2ras(idx:Union[np.ndarray,cp.ndarray], # idx array
           pc_data:Union[np.ndarray,cp.ndarray], # data, 1D or more
           shape:tuple, # image shape
):
    '''convert sparse data to raster, filled with nan'''
    xp = cp.get_array_module(pc_data)
    raster = xp.empty((*shape,*pc_data.shape[1:]),dtype=pc_data.dtype)
    raster[:] = xp.nan
    raster[idx[0],idx[1]] = pc_data
    return raster

# %% ../nbs/API/pc.ipynb 7
def _ras_dims(idx1:Union[np.ndarray,cp.ndarray], # int array, index of the first point cloud
              idx2:Union[np.ndarray,cp.ndarray], # int array, index of the second point cloud
             )->tuple: # the shape of the original raster image
    '''Get the shape of the original raster image from two index, the shape could be smaller than the truth but it doesn't matter.'''
    xp = cp.get_array_module(idx1)
    dims_az = max(int(idx1[0,-1]),int(idx2[0,-1]))+1
    dims_r = max(int(xp.max(idx1[1,:])),int(xp.max(idx2[1,:])))+1
    return (dims_az,dims_r)

# %% ../nbs/API/pc.ipynb 8
def pc_union(idx1:Union[np.ndarray,cp.ndarray], # int array, index of the first point cloud
             idx2:Union[np.ndarray,cp.ndarray], # int array, index of the second point cloud
            )->tuple: # the union index `idx`; index of the point in output union index that originally in the first point cloud `inv_iidx`; index of the point in output union index that only exist in the second point cloud `inv_iidx2`; index of the point in the second input index that are not in the first input point cloud
    '''Get the union of two point cloud dataset. For points at their intersection, pc_data1 rather than pc_data2 is copied to the result pc_data.'''
    # this function is modified from np.unique

    xp = cp.get_array_module(idx1)
    dims = _ras_dims(idx1,idx2)

    idx = xp.concatenate((idx1,idx2),axis=-1)
    n1 = idx1.shape[1]; n2 = idx2.shape[1]
    dims = (100,100)
    
    idx_1d = xp.ravel_multi_index(idx,dims=dims) # automatically the returned 1d index is in int64
    iidx = xp.argsort(idx_1d,kind='stable') # test shows argsort is faster than lexsort, that is why use ravel and unravel index
    idx_1d = idx_1d[iidx]

    inv_iidx = xp.empty_like(iidx)
    inv_iidx[iidx] = xp.arange(iidx.shape[0]) # idea taken from https://stackoverflow.com/questions/2483696/undo-or-reverse-argsort-python

    mask = xp.empty(idx_1d.shape, dtype=bool)
    mask[:1] = True
    mask[1:] = idx_1d[1:] != idx_1d[:-1]
    
    idx_1d = idx_1d[mask]
    
    _mask = mask[inv_iidx] # the mask in the original cat order
    mask1 = _mask[:n1]
    mask2 = _mask[n1:]
    
    imask = xp.cumsum(mask) - 1
    inv_iidx = xp.empty(mask.shape, dtype=np.int64)
    inv_iidx[iidx] = imask # inverse the mapping
    inv_iidx = inv_iidx[_mask]
    
    idx = xp.stack(xp.unravel_index(idx_1d,dims)).astype(idx1.dtype)
   
    return idx, inv_iidx[:n1], inv_iidx[n1:], *xp.where(mask2)

# %% ../nbs/API/pc.ipynb 15
def pc_intersect(idx1:Union[np.ndarray,cp.ndarray], # int array, index of the first point cloud
                 idx2:Union[np.ndarray,cp.ndarray], # int array, index of the second point cloud
                 # the intersect index `idx`,
                 # index of the point in first point cloud index that also exist in the second point cloud,
                 # index of the point in second point cloud index that also exist in the first point cloud
                )->tuple:
    '''Get the intersection of two point cloud dataset.'''
    # Here I do not write the core function by myself since cupy have a different implementation of intersect1d

    xp = cp.get_array_module(idx1)
    dims = _ras_dims(idx1,idx2)

    idx1_1d = xp.ravel_multi_index(idx1,dims=dims) # automatically the returned 1d index is in int64
    idx2_1d = xp.ravel_multi_index(idx2,dims=dims) # automatically the returned 1d index is in int64

    idx, iidx1, iidx2 = xp.intersect1d(idx1_1d,idx2_1d,assume_unique=True,return_indices=True)
    idx = xp.stack(xp.unravel_index(idx,dims)).astype(idx1.dtype)

    return idx, iidx1, iidx2

# %% ../nbs/API/pc.ipynb 17
def pc_diff(idx1:Union[np.ndarray,cp.ndarray], # int array, index of the first point cloud
            idx2:Union[np.ndarray,cp.ndarray], # int array, index of the second point cloud
            # the diff index `idx`,
            # index of the point in first point cloud index that do not exist in the second point cloud,
           )->tuple:
    '''Get the point cloud in `idx1` that are not in `idx2`.'''

    xp = cp.get_array_module(idx1)
    dims = _ras_dims(idx1,idx2)

    idx1_1d = xp.ravel_multi_index(idx1,dims=dims) # automatically the returned 1d index is in int64
    idx2_1d = xp.ravel_multi_index(idx2,dims=dims) # automatically the returned 1d index is in int64
    
    mask = xp.in1d(idx1_1d, idx2_1d, assume_unique=True, invert=True)
    idx = idx1_1d[mask]

    idx = xp.stack(xp.unravel_index(idx,dims)).astype(idx1.dtype)
    return idx, xp.where(mask)[0]



# AUTOGENERATED! DO NOT EDIT! File to edit: ../../../nbs/CLI/utils/logging.ipynb.

# %% auto 0
__all__ = ['zarr_info', 'darr_info', 'get_logger', 'log_args']

# %% ../../../nbs/CLI/utils/logging.ipynb 3
import logging
import sys
import random
import string
import inspect
from functools import wraps
import types

import zarr
from dask import array as da

# %% ../../../nbs/CLI/utils/logging.ipynb 4
def _get_random_string(length=36):
    # choose from all lowercase letter
    letters = string.ascii_lowercase
    result_str = ''.join(random.choice(letters) for i in range(length))
    return result_str

# %% ../../../nbs/CLI/utils/logging.ipynb 5
def zarr_info(self, # logger
              path, # string to zarr
              zarr, # zarr dataset
             ):
    self.info(f'{path} zarray shape: '+str(zarr.shape))
    self.info(f'{path} zarray chunks: '+str(zarr.chunks))
    self.info(f'{path} zarray dtype: '+str(zarr.dtype))

# %% ../../../nbs/CLI/utils/logging.ipynb 6
def darr_info(self, # logger
              name, # printing name of the dask array
              darr, # dask array
             ):
    self.info(f'{name} dask array shape: '+str(darr.shape))
    self.info(f'{name} dask array chunksize: '+str(darr.chunksize))
    self.info(f'{name} dask array dtype: '+str(darr.dtype))

# %% ../../../nbs/CLI/utils/logging.ipynb 7
def get_logger(name:str=None, # name of the application, optional. default: the function name that call this function
               logfile:str=None, # logfile, optional. default: no logfile
               level:str=None, # log level, debug or info, optional. default: info
              ):
    '''get logger for decorrelation cli application'''

    if not name:
        name = inspect.stack()[1][3] #obtain the previous level function name
    
    if not level:
        level = 'info'
    if level == 'info':
        level = logging.INFO
    elif level == 'debug':
        level = logging.DEBUG
    else:
        raise NotImplementedError('only debug and info level are supported')

    LOG_FORMAT = "%(asctime)s - %(levelname)s - %(message)s"
    DATE_FORMAT = "%m/%d/%Y %H:%M:%S %p"
    random_logger_name = _get_random_string(36)
    #The real name of the logger is set to random string to prevent events propogating.

    logger = logging.getLogger(random_logger_name)
    logger.zarr_info = types.MethodType(zarr_info, logger)
    logger.darr_info = types.MethodType(darr_info, logger)

    logger.setLevel(level)
    formatter = logging.Formatter(f'%(asctime)s - {name} - %(levelname)s - %(message)s',datefmt='%Y-%m-%d %H:%M:%S')

    console_handler = logging.StreamHandler(sys.stdout)
    console_handler.setLevel(level)
    console_handler.setFormatter(formatter)
    logger.addHandler(console_handler)
    
    if logfile:
        file_handler = logging.FileHandler(logfile)
        file_handler.setLevel(level)
        file_handler.setFormatter(formatter)
        logger.addHandler(file_handler)
    return logger

# %% ../../../nbs/CLI/utils/logging.ipynb 8
def log_args(func):
    @wraps(func)
    def wrapper(*args, **kwargs):
        ba = inspect.signature(func).bind(*args, **kwargs)
        ba.apply_defaults()
        func_args = ba.arguments
        logger = get_logger(func.__name__,logfile=func_args['log'])
        func_args_strs = map("{0[0]} = {0[1]!r}".format, func_args.items())
        logger.info('fetching args:')
        for item in func_args_strs:
            logger.info(item)
        logger.info('fetching args done.')
        return func(*args, **kwargs)
    return wrapper



# Autogenerated by nbdev

d = { 'settings': { 'branch': 'main',
                'doc_baseurl': '/decorrelation',
                'doc_host': 'https://kanglcn.github.io',
                'git_url': 'https://github.com/kanglcn/decorrelation',
                'lib_path': 'decorrelation'},
  'syms': { 'decorrelation.cli.co': { 'decorrelation.cli.co.de_emperical_co_pc': ( 'CLI/co.html#de_emperical_co_pc',
                                                                                   'decorrelation/cli/co.py')},
            'decorrelation.cli.load': { 'decorrelation.cli.load._cor_pos_dem': ('CLI/load.html#_cor_pos_dem', 'decorrelation/cli/load.py'),
                                        'decorrelation.cli.load._fetch_slc_par_date': ( 'CLI/load.html#_fetch_slc_par_date',
                                                                                        'decorrelation/cli/load.py'),
                                        'decorrelation.cli.load._geo_width_nlines': ( 'CLI/load.html#_geo_width_nlines',
                                                                                      'decorrelation/cli/load.py'),
                                        'decorrelation.cli.load._rdc_width_nlines': ( 'CLI/load.html#_rdc_width_nlines',
                                                                                      'decorrelation/cli/load.py'),
                                        'decorrelation.cli.load.de_load_gamma_flatten_rslc': ( 'CLI/load.html#de_load_gamma_flatten_rslc',
                                                                                               'decorrelation/cli/load.py'),
                                        'decorrelation.cli.load.de_load_gamma_lat_lon_hgt': ( 'CLI/load.html#de_load_gamma_lat_lon_hgt',
                                                                                              'decorrelation/cli/load.py'),
                                        'decorrelation.cli.load.de_load_gamma_look_vector': ( 'CLI/load.html#de_load_gamma_look_vector',
                                                                                              'decorrelation/cli/load.py'),
                                        'decorrelation.cli.load.de_load_gamma_metadata': ( 'CLI/load.html#de_load_gamma_metadata',
                                                                                           'decorrelation/cli/load.py'),
                                        'decorrelation.cli.load.de_load_gamma_range': ( 'CLI/load.html#de_load_gamma_range',
                                                                                        'decorrelation/cli/load.py'),
                                        'decorrelation.cli.load.read_gamma_image': ( 'CLI/load.html#read_gamma_image',
                                                                                     'decorrelation/cli/load.py'),
                                        'decorrelation.cli.load.read_gamma_plist': ( 'CLI/load.html#read_gamma_plist',
                                                                                     'decorrelation/cli/load.py'),
                                        'decorrelation.cli.load.write_gamma_image': ( 'CLI/load.html#write_gamma_image',
                                                                                      'decorrelation/cli/load.py'),
                                        'decorrelation.cli.load.write_gamma_plist': ( 'CLI/load.html#write_gamma_plist',
                                                                                      'decorrelation/cli/load.py')},
            'decorrelation.cli.pc': { 'decorrelation.cli.pc.console_de_pc2ras': ( 'CLI/pc.html#console_de_pc2ras',
                                                                                  'decorrelation/cli/pc.py'),
                                      'decorrelation.cli.pc.console_de_ras2pc': ( 'CLI/pc.html#console_de_ras2pc',
                                                                                  'decorrelation/cli/pc.py'),
                                      'decorrelation.cli.pc.de_pc2ras': ('CLI/pc.html#de_pc2ras', 'decorrelation/cli/pc.py'),
                                      'decorrelation.cli.pc.de_pc_diff': ('CLI/pc.html#de_pc_diff', 'decorrelation/cli/pc.py'),
                                      'decorrelation.cli.pc.de_pc_intersect': ('CLI/pc.html#de_pc_intersect', 'decorrelation/cli/pc.py'),
                                      'decorrelation.cli.pc.de_pc_select_data': ( 'CLI/pc.html#de_pc_select_data',
                                                                                  'decorrelation/cli/pc.py'),
                                      'decorrelation.cli.pc.de_pc_thres_pc': ('CLI/pc.html#de_pc_thres_pc', 'decorrelation/cli/pc.py'),
                                      'decorrelation.cli.pc.de_pc_thres_ras': ('CLI/pc.html#de_pc_thres_ras', 'decorrelation/cli/pc.py'),
                                      'decorrelation.cli.pc.de_pc_union': ('CLI/pc.html#de_pc_union', 'decorrelation/cli/pc.py'),
                                      'decorrelation.cli.pc.de_ras2pc': ('CLI/pc.html#de_ras2pc', 'decorrelation/cli/pc.py')},
            'decorrelation.cli.pl': {'decorrelation.cli.pl.de_emi': ('CLI/pl.html#de_emi', 'decorrelation/cli/pl.py')},
            'decorrelation.cli.shp': { 'decorrelation.cli.shp.de_select_ds_can': ( 'CLI/shp.html#de_select_ds_can',
                                                                                   'decorrelation/cli/shp.py'),
                                       'decorrelation.cli.shp.de_shp_test': ('CLI/shp.html#de_shp_test', 'decorrelation/cli/shp.py')},
            'decorrelation.cli.utils.appconfig': { 'decorrelation.cli.utils.appconfig.get_config': ( 'CLI/utils/appconfig.html#get_config',
                                                                                                     'decorrelation/cli/utils/appconfig.py')},
            'decorrelation.cli.utils.chunk_size': { 'decorrelation.cli.utils.chunk_size.get_az_chunk_size_from_az_chunk_size': ( 'CLI/utils/chunk_size.html#get_az_chunk_size_from_az_chunk_size',
                                                                                                                                 'decorrelation/cli/utils/chunk_size.py'),
                                                    'decorrelation.cli.utils.chunk_size.get_az_chunk_size_from_n_pc_chunk': ( 'CLI/utils/chunk_size.html#get_az_chunk_size_from_n_pc_chunk',
                                                                                                                              'decorrelation/cli/utils/chunk_size.py'),
                                                    'decorrelation.cli.utils.chunk_size.get_pc_chunk_size_from_n_az_chunk': ( 'CLI/utils/chunk_size.html#get_pc_chunk_size_from_n_az_chunk',
                                                                                                                              'decorrelation/cli/utils/chunk_size.py'),
                                                    'decorrelation.cli.utils.chunk_size.get_pc_chunk_size_from_n_pc_chunk': ( 'CLI/utils/chunk_size.html#get_pc_chunk_size_from_n_pc_chunk',
                                                                                                                              'decorrelation/cli/utils/chunk_size.py'),
                                                    'decorrelation.cli.utils.chunk_size.get_pc_chunk_size_from_pc_chunk_size': ( 'CLI/utils/chunk_size.html#get_pc_chunk_size_from_pc_chunk_size',
                                                                                                                                 'decorrelation/cli/utils/chunk_size.py')},
            'decorrelation.cli.utils.dask': { 'decorrelation.cli.utils.dask._pad_chunks': ( 'CLI/utils/dask.html#_pad_chunks',
                                                                                            'decorrelation/cli/utils/dask.py'),
                                              'decorrelation.cli.utils.dask.get_cuda_cluster': ( 'CLI/utils/dask.html#get_cuda_cluster',
                                                                                                 'decorrelation/cli/utils/dask.py'),
                                              'decorrelation.cli.utils.dask.get_pc_chunk_size': ( 'CLI/utils/dask.html#get_pc_chunk_size',
                                                                                                  'decorrelation/cli/utils/dask.py'),
                                              'decorrelation.cli.utils.dask.pad_internal': ( 'CLI/utils/dask.html#pad_internal',
                                                                                             'decorrelation/cli/utils/dask.py')},
            'decorrelation.cli.utils.logging': { 'decorrelation.cli.utils.logging._get_random_string': ( 'CLI/utils/logging.html#_get_random_string',
                                                                                                         'decorrelation/cli/utils/logging.py'),
                                                 'decorrelation.cli.utils.logging.darr_info': ( 'CLI/utils/logging.html#darr_info',
                                                                                                'decorrelation/cli/utils/logging.py'),
                                                 'decorrelation.cli.utils.logging.get_logger': ( 'CLI/utils/logging.html#get_logger',
                                                                                                 'decorrelation/cli/utils/logging.py'),
                                                 'decorrelation.cli.utils.logging.log_args': ( 'CLI/utils/logging.html#log_args',
                                                                                               'decorrelation/cli/utils/logging.py'),
                                                 'decorrelation.cli.utils.logging.zarr_info': ( 'CLI/utils/logging.html#zarr_info',
                                                                                                'decorrelation/cli/utils/logging.py')},
            'decorrelation.co': { 'decorrelation.co.emperical_co': ('API/co.html#emperical_co', 'decorrelation/co.py'),
                                  'decorrelation.co.emperical_co_pc': ('API/co.html#emperical_co_pc', 'decorrelation/co.py'),
                                  'decorrelation.co.isPD': ('API/co.html#ispd', 'decorrelation/co.py'),
                                  'decorrelation.co.nearestPD': ('API/co.html#nearestpd', 'decorrelation/co.py'),
                                  'decorrelation.co.regularize_spectral': ('API/co.html#regularize_spectral', 'decorrelation/co.py')},
            'decorrelation.pc': { 'decorrelation.pc._ras_dims': ('API/pc.html#_ras_dims', 'decorrelation/pc.py'),
                                  'decorrelation.pc.pc2ras': ('API/pc.html#pc2ras', 'decorrelation/pc.py'),
                                  'decorrelation.pc.pc_diff': ('API/pc.html#pc_diff', 'decorrelation/pc.py'),
                                  'decorrelation.pc.pc_intersect': ('API/pc.html#pc_intersect', 'decorrelation/pc.py'),
                                  'decorrelation.pc.pc_union': ('API/pc.html#pc_union', 'decorrelation/pc.py')},
            'decorrelation.pl': { 'decorrelation.pl.emi': ('API/pl.html#emi', 'decorrelation/pl.py'),
                                  'decorrelation.pl.temp_coh': ('API/pl.html#temp_coh', 'decorrelation/pl.py')},
            'decorrelation.plot': {'decorrelation.plot.bg_alpha': ('API/plot.html#bg_alpha', 'decorrelation/plot.py')},
            'decorrelation.shp': {'decorrelation.shp.ks_test': ('API/shp.html#ks_test', 'decorrelation/shp.py')}}}



# AUTOGENERATED! DO NOT EDIT! File to edit: ../../nbs/CLI/pc.ipynb.

# %% auto 0
__all__ = ['de_ras2pc', 'console_de_ras2pc', 'de_pc2ras', 'console_de_pc2ras', 'de_pc_union', 'de_pc_intersect', 'de_pc_diff',
           'de_pc_thres_ras', 'de_pc_thres_pc', 'de_pc_select_data']

# %% ../../nbs/CLI/pc.ipynb 3
import math
import zarr
import cupy as cp
import numpy as np
from matplotlib import pyplot as plt

import dask
from dask import array as da
from dask.distributed import Client, LocalCluster

from ..pc import pc2ras, pc_union, pc_intersect, pc_diff
from .utils.logging import get_logger, log_args
from .utils.chunk_size import (get_pc_chunk_size_from_n_pc_chunk, 
                                                get_pc_chunk_size_from_pc_chunk_size, 
                                                get_pc_chunk_size_from_n_az_chunk,
                                                get_az_chunk_size_from_n_pc_chunk)

from fastcore.script import call_parse, Param

# %% ../../nbs/CLI/pc.ipynb 4
@log_args
def de_ras2pc(idx:str, # point cloud index
              ras:str|list, # path (in string) or list of path for raster data
              pc:str|list, # output, path (in string) or list of path for point cloud data
              pc_chunk_size:int=None, # output point chunk size
              n_pc_chunk:int=None, # output number of chunk
              hd_chunk_size:tuple|list=None, # output high dimension chunk size, tuple or list of tuple, same as input raster data by default
              log:str=None, # log file. Default: no log file
):
    '''Convert raster data to point cloud data'''
    logger = get_logger(logfile=log)

    idx_zarr = zarr.open(idx,mode='r')
    logger.zarr_info(idx,idx_zarr)
    assert idx_zarr.ndim == 2, "idx dimentation is not 2."
    pc_chunk_size = get_pc_chunk_size_from_pc_chunk_size('idx','pc',idx_zarr.chunks[1],idx_zarr.shape[1],logger,pc_chunk_size=pc_chunk_size,n_pc_chunk=n_pc_chunk)

    logger.info('loading idx into memory.')
    idx = zarr.open(idx,mode='r')[:]
    n_pc = idx.shape[1]

    if isinstance(ras,str):
        assert isinstance(pc,str)
        ras_list = [ras]; pc_list = [pc]
        if hd_chunk_size is not None:
            assert isinstance(hd_chunk_size,tuple)
            hd_chunk_size_list = [hd_chunk_size]
        else:
            hd_chunk_size_list = [None]
    else:
        assert isinstance(ras,list); assert isinstance(pc,list)
        ras_list = ras; pc_list = pc
        n_data = len(ras_list)
        if hd_chunk_size is not None:
            assert isinstance(hd_chunk_size,list)
            hd_chunk_size_list = hd_chunk_size
        else:
            hd_chunk_size_list = [None]*n_data

    logger.info('starting dask local cluster.')
    cluster = LocalCluster()
    client = Client(cluster)
    logger.info('dask local cluster started.')

    _pc_list = ()
    for ras_path, pc_path, hd_chunk_size in zip(ras_list,pc_list,hd_chunk_size_list):
        logger.info(f'start to slice on {ras_path}')
        ras_zarr = zarr.open(ras_path,'r'); logger.zarr_info(ras_path, ras_zarr)
        if hd_chunk_size is None:
            logger.info(f'hd_chunk_size not setted. Use the one from {ras_path}.')
            hd_chunk_size = ras_zarr.chunks[2:]
        logger.info(f'hd_chunk_size: {hd_chunk_size}.')

        ras = da.from_zarr(ras_path,chunks=(*ras_zarr.chunks[:2],*hd_chunk_size)); logger.darr_info('ras',ras)

        with dask.config.set(**{'array.slicing.split_large_chunks': False}):
            pc = ras.reshape(-1,*ras.shape[2:])[np.ravel_multi_index((idx[0],idx[1]),dims=ras.shape[:2])]
        
        logger.darr_info('pc', pc)
        logger.info('rechunk pc data:')
        pc = pc.rechunk((pc_chunk_size,*pc.chunksize[1:]))
        logger.darr_info('pc', pc)
        _pc = pc.to_zarr(pc_path,overwrite=True,compute=False)
        logger.info(f'saving to {pc_path}.')
        _pc_list += (_pc,)
    
    logger.info('computing graph setted. doing all the computing.')
    da.compute(*_pc_list)

    logger.info('computing finished.')
    cluster.close()
    logger.info('dask cluster closed.')

# %% ../../nbs/CLI/pc.ipynb 5
@call_parse
def console_de_ras2pc(idx:str, # point cloud index
                      ras:Param(type=str,required=True,nargs='+',help='one or more path for raster data')=None,
                      pc:Param(type=str,required=True,nargs='+',help='output, one or more path for point cloud data')=None,
                      pc_chunk_size:int=None, # output point chunk size, same as input idx by default
                      n_pc_chunk:int=None, # output number of chunk
                      hd_chunk_size:Param(type=str,nargs='+',help='''output high dimension chunk size,
                      each size should be wrapped in quotation marks and size in each dimension are seperated with ",",
                      same as input raster data by default''')=None,
                      log:str=None, # log file. Default: no log file
):
    '''Convert raster data to point cloud data'''
    if hd_chunk_size is not None:
        hd_chunk_size_ = []
        for size in hd_chunk_size:
            if len(size) == 0:
                size = ()
            else:
                size = size.split(',')
                size = tuple([int(i) for i in size])
            hd_chunk_size_.append(size)
    else:
        hd_chunk_size_ = None

    if len(ras)==1:
        ras = ras[0]
        pc = pc[0]
        if hd_chunk_size_ is not None:
            hd_chunk_size_ = hd_chunk_size_[0]

    de_ras2pc(idx,ras,pc,pc_chunk_size=pc_chunk_size,n_pc_chunk=n_pc_chunk,hd_chunk_size=hd_chunk_size_,log=log)

# %% ../../nbs/CLI/pc.ipynb 10
@log_args
def de_pc2ras(idx:str, # point cloud index
              pc:str|list, # path (in string) or list of path for point cloud data
              ras:str|list, # output, path (in string) or list of path for raster data
              shape:tuple, # shape of one image (nlines,width)
              az_chunk_size:int=None, # output azimuth chunk size, 
              n_az_chunk:int=None, # # output number of azimuth chunks 
              log:str=None, # log file. Default: no log file
):
    '''Convert point cloud data to raster data, filled with nan'''
    logger = get_logger(logfile=log)

    idx_zarr = zarr.open(idx,mode='r')
    logger.info('idx dataset shape: '+str(idx_zarr.shape))
    logger.info('idx dataset chunks: '+str(idx_zarr.chunks))
    assert idx_zarr.ndim == 2, "idx dimentation is not 2."
    az_chunk_size = get_az_chunk_size_from_n_pc_chunk('idx','ras',idx_zarr.shape[1],idx_zarr.chunks[1],shape[0],logger=logger,az_chunk_size=az_chunk_size,n_az_chunk=n_az_chunk)

    logger.info('loading idx into memory.')
    idx = zarr.open(idx,mode='r')[:]
    n_pc = idx.shape[1]
    
    if isinstance(pc,str):
        assert isinstance(ras,str)
        pc_list = [pc]; ras_list = [ras]
    else:
        assert isinstance(pc,list); assert isinstance(ras,list)
        pc_list = pc; ras_list = ras

    logger.info('starting dask local cluster.')
    cluster = LocalCluster()
    client = Client(cluster)
    logger.info('dask local cluster started.')

    _ras_list = ()

    for ras_path, pc_path in zip(ras_list,pc_list):
        logger.info(f'start to work on {pc_path}')
        pc_zarr = zarr.open(pc_path,'r')
        logger.zarr_info(pc_path,pc_zarr)
        
        pc = da.from_zarr(pc_path)
        logger.darr_info('pc', pc)
        ras = da.empty((shape[0]*shape[1],*pc.shape[1:]),chunks = (az_chunk_size*shape[1],*pc_zarr.chunks[1:]), dtype=pc.dtype)
        ras[:] = np.nan
        ras[np.ravel_multi_index((idx[0],idx[1]),dims=shape)] = pc
        ras = ras.reshape(*shape,*pc.shape[1:])
        logger.info('create ras dask array')
        logger.darr_info('ras', ras)
        _ras = ras.to_zarr(ras_path,overwrite=True,compute=False)
        _ras_list += (_ras,)

    logger.info('computing graph setted. doing all the computing.')
    da.compute(*_ras_list)

    logger.info('computing finished.')
    cluster.close()
    logger.info('dask cluster closed.')

# %% ../../nbs/CLI/pc.ipynb 11
@call_parse
def console_de_pc2ras(idx:str, # point cloud index
                      pc:Param(type=str,required=True,nargs='+',help='one or more path for point cloud data')=None,
                      ras:Param(type=str,required=True,nargs='+',help='output, one or more path for raster data')=None,
                      shape:Param(type=str,required=True,help='shape of one image "nlines,width"')=None,
                      az_chunk_size:int=None, # output azimuth chunk size
                      n_az_chunk:int=None, # output number of azimuth chunks 
                      log:str=None, # log file. Default: no log file
):
    '''Convert point cloud data to raster data'''
    if len(ras)==1:
        ras = ras[0]
        pc = pc[0]
    
    shape = shape.split(',')
    shape = [int(i) for i in shape]
    shape=tuple(shape)

    de_pc2ras(idx,pc,ras,shape,az_chunk_size=az_chunk_size,n_az_chunk=n_az_chunk,log=log)

# %% ../../nbs/CLI/pc.ipynb 16
@call_parse
@log_args
def de_pc_union(idx1:str, # index of the first point cloud
                idx2:str, # index of the second point cloud
                idx:str, # output, index of the union point cloud
                pc1:str|list=None, # path (in string) or list of path for the first point cloud data
                pc2:str|list=None, # path (in string) or list of path for the second point cloud data
                pc:str|list=None, #output, path (in string) or list of path for the union point cloud data
                pc_chunk_size:int=None, # chunk size in output data,optional
                n_pc_chunk:int=None, # number of chunk in output data, optional
                log:str=None, # log file. Default: no log file
):
    '''Get the union of two point cloud dataset.
    For points at their intersection, pc_data1 rather than pc_data2 is copied to the result pc_data.
    `pc_chunk_size` and `n_pc_chunk` are used to determine the final pc_chunk_size.
    If non of them are provided, the n_pc_chunk is set to n_chunk in idx1.
    '''
    logger = get_logger(logfile=log)

    idx1_zarr = zarr.open(idx1,mode='r'); logger.zarr_info(idx1,idx1_zarr)
    idx2_zarr = zarr.open(idx2,mode='r'); logger.zarr_info(idx2,idx2_zarr)
    logger.info('loading idx1 and idx2 into memory.')
    idx1 = idx1_zarr[:]; idx2 = idx2_zarr[:]

    logger.info('calculate the union')
    idx_path = idx
    idx, inv_iidx1, inv_iidx2, iidx2 = pc_union(idx1,idx2)
    n_pc = idx.shape[1]
    logger.info(f'number of points in the union: {idx.shape[1]}')
    pc_chunk_size = get_pc_chunk_size_from_n_pc_chunk('idx1','idx',idx1_zarr.shape[1],idx1_zarr.chunks[1],n_pc,logger,pc_chunk_size=pc_chunk_size,n_pc_chunk=n_pc_chunk)
    
    idx_zarr = zarr.open(idx_path,'w',shape=idx.shape,dtype=idx.dtype,chunks=(2,pc_chunk_size))
    logger.info('write union idx')
    idx_zarr[:] = idx
    logger.info('write done')
    logger.zarr_info(idx_path, idx_zarr)
    
    if pc1 is None:
        logger.info('no point cloud data provided, exit.')
        return None

    if isinstance(pc1,str):
        assert isinstance(pc2,str); assert isinstance(pc,str)
        pc1_list = [pc1]; pc2_list = [pc2]; pc_list = [pc]
    else:
        assert isinstance(pc1,list); assert isinstance(pc2,list); assert isinstance(pc,list)
        pc1_list = pc1; pc2_list = pc2; pc_list = pc

    logger.info('starting dask local cluster.')
    cluster = LocalCluster()
    client = Client(cluster)
    logger.info('dask local cluster started.')
    
    _pc_list = ()
    for pc1_path, pc2_path, pc_path in zip(pc1_list,pc2_list,pc_list):
        pc1_zarr = zarr.open(pc1_path,'r'); pc2_zarr = zarr.open(pc2_path,'r')
        logger.zarr_info(pc1_path, pc1_zarr); logger.zarr_info(pc2_path, pc2_zarr);
        pc1 = da.from_zarr(pc1_path); pc2 = da.from_zarr(pc2_path)
        logger.darr_info('pc1', pc1); logger.darr_info('pc2',pc2)
        logger.info('set up union pc data dask array.')
        pc = da.empty((n_pc,*pc1.shape[1:]),chunks = (pc_chunk_size,*pc1.chunks[1:]), dtype=pc1.dtype)
        logger.darr_info('pc',pc)
        pc[inv_iidx1] = pc1
        pc[inv_iidx2] = pc2[iidx2]
        _pc = pc.to_zarr(pc_path, overwrite=True,compute=False)
        _pc_list += (_pc,)

    logger.info('computing graph setted. doing all the computing.')
    da.compute(*_pc_list)

    logger.info('computing finished.')
    cluster.close()
    logger.info('dask cluster closed.')

# %% ../../nbs/CLI/pc.ipynb 20
@call_parse
@log_args
def de_pc_intersect(idx1:str, # index of the first point cloud
                    idx2:str, # index of the second point cloud
                    idx:str, # output, index of the union point cloud
                    pc1:str|list=None, # path (in string) or list of path for the first point cloud data
                    pc2:str|list=None, # path (in string) or list of path for the second point cloud data
                    pc:str|list=None, #output, path (in string) or list of path for the union point cloud data
                    pc_chunk_size:int=None, # chunk size in output data,optional
                    n_pc_chunk:int=None, # number of chunk in output data, optional
                    prefer_1=True, # save pc1 on intersection to output pc dataset by default `True`. Otherwise, save data from pc2
                    log:str=None, # log file. Default: no log file
):
    '''Get the intersection of two point cloud dataset.
    `pc_chunk_size` and `n_pc_chunk` are used to determine the final pc_chunk_size.
    If non of them are provided, the n_pc_chunk is set to n_chunk in idx1.
    '''
    logger = get_logger(logfile=log)

    idx1_zarr = zarr.open(idx1,mode='r'); logger.zarr_info(idx1,idx1_zarr)
    idx2_zarr = zarr.open(idx2,mode='r'); logger.zarr_info(idx2,idx2_zarr)
    logger.info('loading idx1 and idx2 into memory.')
    idx1 = idx1_zarr[:]; idx2 = idx2_zarr[:]

    logger.info('calculate the intersection')
    idx_path = idx
    idx, iidx1, iidx2 = pc_intersect(idx1,idx2)
    n_pc = idx.shape[1]
    logger.info(f'number of points in the intersection: {idx.shape[1]}')
    pc_chunk_size = get_pc_chunk_size_from_n_pc_chunk('idx1','idx',idx1_zarr.shape[1],idx1_zarr.chunks[1],n_pc,logger,pc_chunk_size=pc_chunk_size,n_pc_chunk=n_pc_chunk)
    
    idx_zarr = zarr.open(idx_path,'w',shape=idx.shape,dtype=idx.dtype,chunks=(2,pc_chunk_size))
    logger.info('write intersect idx')
    idx_zarr[:] = idx
    logger.info('write done')
    logger.zarr_info(idx_path, idx_zarr)

    if (pc1 is None) and (pc2 is None):
        logger.info('no point cloud data provided, exit.')
        return None

    if prefer_1:
        logger.info('select pc1 as pc_input.')
        iidx = iidx1; pc_input = pc1
    else:
        logger.info('select pc2 as pc_input.')
        iidx = iidx2; pc_input = pc2

    if isinstance(pc_input,str):
        assert isinstance(pc,str)
        pc_input_list = [pc_input]; pc_list = [pc]
    else:
        assert isinstance(pc_input,list); assert isinstance(pc,list)
        pc_input_list = pc_input; pc_list = pc

    logger.info('starting dask local cluster.')
    cluster = LocalCluster(); client = Client(cluster)
    logger.info('dask local cluster started.')
    
    _pc_list = ()
    for pc_input_path, pc_path in zip(pc_input_list,pc_list):
        pc_input_zarr = zarr.open(pc_input_path,'r')
        logger.zarr_info(pc_input_path,pc_input_zarr)
        pc_input = da.from_zarr(pc_input_path)
        logger.darr_info('pc_input', pc_input)

        logger.info('set up intersect pc data dask array.')
        pc = da.empty((n_pc,*pc_input.shape[1:]),chunks = (pc_chunk_size,*pc_input.chunks[1:]), dtype=pc_input.dtype)
        logger.darr_info('pc',pc)
        pc[:] = pc_input[iidx]
        _pc = pc.to_zarr(pc_path, overwrite=True,compute=False)
        _pc_list += (_pc,)

    logger.info('computing graph setted. doing all the computing.')
    da.compute(*_pc_list)

    logger.info('computing finished.')
    cluster.close()
    logger.info('dask cluster closed.')

# %% ../../nbs/CLI/pc.ipynb 24
@call_parse
@log_args
def de_pc_diff(idx1:str, # index of the first point cloud
               idx2:str, # index of the second point cloud
               idx:str, # output, index of the union point cloud
               pc1:str|list=None, # path (in string) or list of path for the first point cloud data
               pc:str|list=None, #output, path (in string) or list of path for the union point cloud data
               pc_chunk_size:int=None, # chunk size in output data,optional
               n_pc_chunk:int=None, # number of chunk in output data, optional
               log:str=None, # log file. Default: no log file
              ):
    '''Get the point cloud in `idx1` that are not in `idx2`.
    `pc_chunk_size` and `n_pc_chunk` are used to determine the final pc_chunk_size.
    If non of them are provided, the n_pc_chunk is set to n_chunk in idx1.
    '''
    logger = get_logger(logfile=log)

    idx1_zarr = zarr.open(idx1,mode='r'); logger.zarr_info(idx1,idx1_zarr)
    idx2_zarr = zarr.open(idx2,mode='r'); logger.zarr_info(idx2,idx2_zarr)
    logger.info('loading idx1 and idx2 into memory.')
    idx1 = idx1_zarr[:]; idx2 = idx2_zarr[:]

    logger.info('calculate the diff.')
    idx_path = idx
    idx, iidx1 = pc_diff(idx1,idx2)
    n_pc = idx.shape[1]
    logger.info(f'number of points in the diff: {idx.shape[1]}')
    pc_chunk_size = get_pc_chunk_size_from_n_pc_chunk('idx1','idx',idx1_zarr.shape[1],idx1_zarr.chunks[1],n_pc,logger,pc_chunk_size=pc_chunk_size,n_pc_chunk=n_pc_chunk)
    
    idx_zarr = zarr.open(idx_path,'w',shape=idx.shape,dtype=idx.dtype,chunks=(2,pc_chunk_size))
    logger.info('write intersect idx')
    idx_zarr[:] = idx
    logger.info('write done')
    logger.zarr_info(idx_path, idx_zarr)

    if pc1 is None:
        logger.info('no point cloud data provided, exit.')
        return None

    if isinstance(pc1,str):
        assert isinstance(pc,str)
        pc1_list = [pc1]; pc_list = [pc]
    else:
        assert isinstance(pc1,list); assert isinstance(pc,list)
        pc1_list = pc1; pc_list = pc

    logger.info('starting dask local cluster.')
    cluster = LocalCluster(); client = Client(cluster)
    logger.info('dask local cluster started.')
    
    _pc_list = ()
    for pc1_path, pc_path in zip(pc1_list,pc_list):
        pc1_zarr = zarr.open(pc1_path,'r'); logger.zarr_info(pc1_path, pc1_zarr)
        pc1 = da.from_zarr(pc1_path); logger.darr_info('pc1', pc1)
        logger.info('set up diff pc data dask array.')
        pc = da.empty((n_pc,*pc1.shape[1:]),chunks = (pc_chunk_size,*pc1.chunks[1:]), dtype=pc1.dtype)
        logger.darr_info('pc',pc)
        pc[:] = pc1[iidx1]
        _pc = pc.to_zarr(pc_path, overwrite=True,compute=False)
        _pc_list += (_pc,)

    logger.info('computing graph setted. doing all the computing.')
    da.compute(*_pc_list)

    logger.info('computing finished.')
    cluster.close()
    logger.info('dask cluster closed.')

# %% ../../nbs/CLI/pc.ipynb 28
@call_parse
@log_args
def de_pc_thres_ras(ras, # the raster image used for thresholding
                    idx, # output, index of selected pixels
                    min_thres=None, # minimum value of `thres_ras` pixels to be selected in the output point cloud
                    max_thres=None, # maximum value of `thres_ras` pixels to be selected in the output point cloud
                    pc_chunk_size:int=None, # chunk size in output data,optional
                    n_pc_chunk:int=None, # number of chunk in output data, optional
                    log:str=None, # log file. Default: no log file
                   ):
    '''generate point cloud index based on threshold of one raster image.
    '''
    idx_path = idx
    logger = get_logger(logfile=log)
    ras_zarr = zarr.open(ras, mode='r'); logger.zarr_info(ras,ras_zarr)

    ras = ras_zarr[:]; logger.info('loading ras into memory.')
    if (min_thres is not None) and (max_thres is not None):
        is_pc = (ras >= min_thres)&(ras<= max_thres)
        logger.info('select pc based on min_thres and max_thres.')
    elif (min_thres is not None) and (max_thres is None):
        is_pc = ras >= min_thres
        logger.info('select pc based on min_thres.')
    elif (min_thres is None) and (max_thres is not None):
        is_pc = ras<= max_thres
        logger.info('select pc based on max_thres.')
    else:
        is_pc = np.ones_like(ras,dtype=bool)
        logger.info('no input min_thres and max_thres, select all pixels')
    idx = np.stack(np.where(is_pc)).astype(np.int32)
    n_pc = idx.shape[1]
    logger.info(f'number of selected pixels: {n_pc}.')
    pc_chunk_size = get_pc_chunk_size_from_n_az_chunk('ras','idx',ras_zarr.shape[0],ras_zarr.chunks[0],n_pc,logger=logger,pc_chunk_size=pc_chunk_size,n_pc_chunk=n_pc_chunk)
    idx_zarr = zarr.open(idx_path,'w',dtype=idx.dtype,shape=idx.shape,chunks=(2,pc_chunk_size))
    logger.info('writing idx.')
    idx_zarr[:] = idx

# %% ../../nbs/CLI/pc.ipynb 31
@call_parse
@log_args
def de_pc_thres_pc(idx_in,# the index of input pc data
                   pc_in, # the point cloud data used for thresholding
                   idx, # output, index of selected pixels
                   min_thres=None, # minimum value of `thres_ras` pixels to be selected in the output point cloud
                   max_thres=None, # maximum value of `thres_ras` pixels to be selected in the output point cloud
                   pc_chunk_size:int=None, # chunk size in output data,optional
                   n_pc_chunk:int=None, # number of chunk in output data, optional
                   log:str=None, # log file. Default: no log file
                   ):
    '''generate point cloud index and data based on threshold of one point cloud data.
    '''
    idx_path = idx
    logger = get_logger(logfile=log)
    idx_in_zarr = zarr.open(idx_in,mode='r'); logger.zarr_info(idx_in,idx_in_zarr)
    pc_in_zarr = zarr.open(pc_in, mode='r'); logger.zarr_info(pc_in,pc_in_zarr)

    idx_in = idx_in_zarr[:]; logger.info('loading idx_in into memory.')
    pc_in = pc_in_zarr[:]; logger.info('loading pc_in into memory.')

    if (min_thres is not None) and (max_thres is not None):
        is_pc = (pc_in >= min_thres)&(pc_in <= max_thres)
        logger.info('select pc based on min_thres and max_thres.')
    elif (min_thres is not None) and (max_thres is None):
        is_pc = pc_in >= min_thres
        logger.info('select pc based on min_thres.')
    elif (min_thres is None) and (max_thres is not None):
        is_pc = pc_in <= max_thres
        logger.info('select pc based on max_thres.')
    else:
        is_pc = np.ones_like(pc_in,dtype=bool)
        logger.info('no input min_thres and max_thres, select all pixels')

    idx = idx_in[:,is_pc]
    n_pc = idx.shape[1]
    logger.info(f'number of selected pixels: {n_pc}.')
    pc_chunk_size = get_pc_chunk_size_from_n_pc_chunk('idx_in','idx',idx_in_zarr.shape[1],idx_in_zarr.chunks[1],n_pc, logger, pc_chunk_size=pc_chunk_size, n_pc_chunk= n_pc_chunk)
    idx_zarr = zarr.open(idx_path,'w',dtype=idx.dtype,shape=idx.shape,chunks=(2,pc_chunk_size))
    logger.info('writing idx.')
    idx_zarr[:] = idx

# %% ../../nbs/CLI/pc.ipynb 35
@call_parse
@log_args
def de_pc_select_data(idx_in:str, # index of the input data
                      idx:str, # index of the output data
                      pc_in:str|list, # path (in string) or list of path for the input point cloud data
                      pc:str|list, # path (in string) or list of path for the output point cloud data
                      pc_chunk_size:int=None, # chunk size in output data,optional
                      n_pc_chunk:int=None, # number of chunk in output data, optional
                      log:str=None, # log file. Default: no log file
                     ):
    '''generate point cloud data based on its index and one point cloud data.
    The index of generated point cloud data must in the index of the old one.
    '''
    idx_in_path = idx_in; idx_path = idx
    logger = get_logger(logfile=log)
    idx_in_zarr = zarr.open(idx_in_path,mode='r'); logger.zarr_info(idx_in_path,idx_in_zarr)
    idx_zarr = zarr.open(idx_path,mode='r'); logger.zarr_info(idx_path,idx_zarr)
    logger.info('loading idx_in and idx into memory.')
    idx_in = idx_in_zarr[:]; idx = idx_zarr[:]
    iidx_in, iidx = pc_intersect(idx_in,idx)[1:]
    np.testing.assert_array_equal(iidx,np.arange(iidx.shape[0]),err_msg='idx have points that are not covered by idx_in.')
    n_pc = iidx_in.shape[0]
    pc_chunk_size = get_pc_chunk_size_from_pc_chunk_size('idx','pc',idx_zarr.chunks[1],n_pc,logger,pc_chunk_size=pc_chunk_size,n_pc_chunk=n_pc_chunk)

    if isinstance(pc_in,str):
        assert isinstance(pc,str)
        pc_in_list = [pc_in]; pc_list = [pc]
    else:
        assert isinstance(pc_in,list); assert isinstance(pc,list)
        pc_in_list = pc_in; pc_list = pc

    logger.info('starting dask local cluster.')
    cluster = LocalCluster(); client = Client(cluster)
    logger.info('dask local cluster started.')
    
    _pc_list = ()
    for pc_in_path, pc_path in zip(pc_in_list,pc_list):
        pc_in_zarr = zarr.open(pc_in_path,'r'); logger.zarr_info(pc_in_path, pc_in_zarr)
        pc_in = da.from_zarr(pc_in_path); logger.darr_info('pc_in', pc_in)
        logger.info('set up selected pc data dask array.')
        pc = da.empty((n_pc,*pc_in.shape[1:]),chunks = (pc_chunk_size,*pc_in.chunks[1:]), dtype=pc_in.dtype)
        logger.darr_info('pc',pc)
        pc[:] = pc_in[iidx_in]
        _pc = pc.to_zarr(pc_path, overwrite=True,compute=False)
        _pc_list += (_pc,)

    logger.info('computing graph setted. doing all the computing.')
    da.compute(*_pc_list)

    logger.info('computing finished.')
    cluster.close()
    logger.info('dask cluster closed.')
